// -------------------------------------------------------------------------
//
// Copyright or © or Copr. CNRS, (November 16, 2004)
//
// This software is a library providing tool for bioinformatics.
//
// This software is governed by the CeCILL  license under French law and
// abiding by the rules of distribution of free software.  You can  use,
// modify and/ or redistribute the software under the terms of the CeCILL
// license as circulated by CEA, CNRS and INRIA at the following URL
// "http://www.cecill.info".
//
// As a counterpart to the access to the source code and  rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty  and the software's author,  the holder of the
// economic rights,  and the successive licensors  have only  limited
// liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading,  using,  modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean  that it is complicated to manipulate,  and  that  also
// therefore means  that it is reserved for developers  and  experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and,  more generally, to use and operate it in the
// same conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.
//
// -------------------------------------------------------------------------

/**
 * <p>Description : Baobab core classes.</p>
 * <p>Copyright : Copyright (c) Julien Dutheil</p>
 * @author Julien Dutheil
 */

package TreeEditor;

import gnu.jtools.license.LicenseDisplayDialog;
import gnu.jtools.utils.OperatingSystem;
import gnu.jtools.utils.commandline.Argument;
import gnu.jtools.utils.commandline.ArgumentsList;
import gnu.jtools.utils.commandline.ArgumentsTools;
import gnu.jtools.utils.gui.GraphicsTools;
import java.util.*;
import java.io.*;
import java.awt.*;
import javax.swing.*;
import java.net.*;

public class Baobab {

  public static ResourceBundle message;
  public static Locale locale;
  public static OperatingSystem OS;
  public static boolean defaultLNF;
  public static File path;
  public static ArgumentsList currentArgumentsList;
  public static File configFile, userdefinedConfigFile;
  public static final String[] AVAILABLE_LOCALE = { "en_US" };
  public static final String[] AVAILABLE_OS     = { "Linux", "Unix", "Win32", "MacOS" };

  private static void setArgumentsFromFile(File file) {
    try {
      ArgumentsList argsList = ArgumentsTools.parseArguments(new BufferedReader(new FileReader(file)));
      currentArgumentsList = argsList;
      setArgumentsFromList(argsList);
    } catch(IOException e) {
      System.err.println("Bad configuration file: " + file.getAbsolutePath());
    }
  }

  static void setArgumentsFromList(ArgumentsList argsList) {
    Argument argument;
    String param;

    //Setting up Look N' Feel:
    argument = argsList.get("LookNFeel");
    if (argument != null) {
      param = argument.getValue();
      if (param.equalsIgnoreCase("METAL")) { //Java default Look N'Feel
        defaultLNF = true;
      }
    }

    //Setting up locale:
    argument = argsList.get("Locale");
    if (argument != null) {
      param = argument.getValue();
      StringTokenizer analyse = new StringTokenizer(param, "_");
      if (analyse.countTokens() == 2) {
        locale = new Locale(new String(analyse.nextToken()),
                            new String(analyse.nextToken()));
        message = ResourceBundle.getBundle("TreeEditor/bundle", locale);
      }
      else {
        System.out.println("Bad Locale parameter: " + param);
        System.exit( -1);
      }
    }

    //Setting up preferred path:
    argument = argsList.get("PreferredPath");
    if (argument != null) {
      param = argsList.get("PreferredPath").getValue();
      if (param.equalsIgnoreCase("Install")) {
        //Installation directory
        path = new File(System.getProperty("user.dir"));
      }
      else {
        try {
          path = new File(param);
          if (!path.exists() || !path.isDirectory())
            throw new IOException();
        }
        catch (IOException e) {
          System.out.println("Error in argument 'PreferredPath'. File " +
                             path.getAbsolutePath() +
                             " does not exists or is not a directory.");
          System.exit( -1);
        }
      }
    }

    //Setting up OS:
    argument = argsList.get("OS");
    if (argument != null) {
      param = argsList.get("OS").getValue();
      if (param.equalsIgnoreCase("MACOS")) {
        OS = new OperatingSystem(OperatingSystem.MacOS);
      }
      if (param.equalsIgnoreCase("LINUX")) {
        OS = new OperatingSystem(OperatingSystem.LINUX);
      }
      if (param.equalsIgnoreCase("UNIX")) {
        OS = new OperatingSystem(OperatingSystem.UNIX);
      }
      if (param.equalsIgnoreCase("WIN32")) {
        OS = new OperatingSystem(OperatingSystem.Win32);
      }
      //Else default (LINUX) is used
    }
  }

  private static void createConfigurationFile(File file) throws IOException {
    Object loc = JOptionPane.showInputDialog(null, "Set locale:", "Baobab installation", JOptionPane.QUESTION_MESSAGE, null, AVAILABLE_LOCALE, AVAILABLE_LOCALE[0]);
    if(loc == null) {
      System.out.println("Installation cancelled by user.");
      configFile.delete();
      System.exit(0);
    }
    Object os = JOptionPane.showInputDialog(null, "Set OS:", "Baobab installation", JOptionPane.QUESTION_MESSAGE, null, AVAILABLE_OS, AVAILABLE_OS[0]);
    if(os == null) {
      System.out.println("Installation cancelled by user.");
      configFile.delete();
      System.exit(0);
    }

    PrintWriter pw = new PrintWriter(new FileWriter(file));
    pw.println("# All directories are relative to the Baobab installation directory");
    pw.println("# i.e. directory '..'.");
    pw.println("#");
    pw.println("PreferredPath=install");
    pw.println("OS=" + (String)os);
    //pw.println("LookNFeel=Metal");
    pw.println("Locale=" + (String)loc);
    pw.close();
  }

  public static void error(Component parent, String text) {
    JOptionPane.showMessageDialog(parent, text, Baobab.message.getString("failure"), JOptionPane.ERROR_MESSAGE);
  }

  public static int warning(Component parent, String warningMessage) {
    return JOptionPane.showConfirmDialog(parent, warningMessage, Baobab.message.getString("warning"), JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
  }

  public static String request(Component parent, String text, String defaultValue) {
    String result = JOptionPane.showInputDialog(parent, text, defaultValue);
    return result;
  }

  public static void message(Component parent, String text) {
    JOptionPane.showMessageDialog(parent, text, Baobab.message.getString("Info"), JOptionPane.INFORMATION_MESSAGE);
  }

  public static void main(String[] args) {
    /**
     * Default arguments:
     */
    defaultLNF = false;
    Locale locale = new Locale("en", "US");
    path = null;
    OS = new OperatingSystem(OperatingSystem.LINUX);

    /**
     * Read parameters from arguments:
     */

    //Read the default configuration file:
    File configDir = new File(System.getProperty("user.home"), ".Baobab");
    configFile = new File(configDir, "config.cfg");
    try {
      if (!configDir.exists()) {
        //Display the license agreements:
        URL licenseURL = LicenseDisplayDialog.class.getResource("Licence_CeCILL_V1-US.txt");
        LicenseDisplayDialog ldd = new LicenseDisplayDialog(
            "CeCILL General Public License", true, licenseURL);
        ldd.setVisible(true);
        if(!ldd.userAcceptsTheTermsOfTheLicense()) {
          JOptionPane.showMessageDialog(
              null,
              "You must remove each copy of this\n program files from your harddisk",
              "License declined.",
              JOptionPane.OK_OPTION);
          System.exit(0);
        }

        configDir.mkdir();
        System.out.println("Created directory " + configDir.getAbsolutePath());
      }
      if (!configFile.exists()) {
        configFile.createNewFile();
        createConfigurationFile(configFile);
      }
    } catch(IOException e) {
      System.err.println("Baobab encountered a problem when initializing configuration files, check permissions.");
      System.exit(-1);
    }
    setArgumentsFromFile(configFile);
    userdefinedConfigFile = new File(configDir, "user-config.cfg");
    if(userdefinedConfigFile.exists()) setArgumentsFromFile(userdefinedConfigFile);

    //Overwrites arguments read from command line
    if(args.length > 0) {
      //Read arguments from command line.
      ArgumentsList argsList = ArgumentsTools.parseArguments(args);
      setArgumentsFromList(argsList);
    }

    if(!defaultLNF) {
      try { UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName()); }
      catch (Exception e) { }
    }
    MainInterface mi = new MainInterface(locale, OS);
    GraphicsTools.center(mi);
    mi.setVisible(true);
  }
}
