package bio.dendogram.GUI;

import gnu.jtools.utils.file.format.FormatAndFileChooser;
import gnu.jtools.utils.xmltools.ParserOptionPanel;
import gnu.jtools.utils.xmltools.XMLTools;
import bio.dendogram.*;
import bio.dendogram.xml.*;
import bio.dendogram.io.*;
import java.awt.*;
import java.awt.event.*;
import java.io.File;
import java.io.IOException;
import javax.swing.*;
import javax.swing.filechooser.*;
import java.util.*;

public class TreeIODialog extends FormatAndFileChooser {

  protected class ParserDialog extends JDialog {

    protected int status;

    public ParserDialog(Frame owner, final TreeXMLParserPanel parserPanel) {
      super(owner, "XML Parser panel", true);
      getContentPane().setLayout(new BorderLayout());
      getContentPane().add(new JScrollPane(parserPanel), BorderLayout.CENTER);
      JToolBar toolBar = new JToolBar("Control");
      toolBar.add(new AbstractAction("Ok") {
        public void actionPerformed(ActionEvent e) { parserDialog.setVisible(false); }
      });
      getContentPane().add(toolBar, BorderLayout.SOUTH);
      pack();
      setDefaultCloseOperation(DISPOSE_ON_CLOSE);
      addWindowListener(new WindowAdapter() {
        public void windowClosing(WindowEvent e) { status = JFileChooser.CANCEL_OPTION; }
      });

      writeListener = new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          if (getSelectedFormat().equals("XML")) {
            try {
              parserPanel.setAvailableTags(xmlIO.getAvailableTags(getSelectedFile()));
            } catch (IOException ioe) {
              // Do nothing, file currently selected is not valid.
            }
          }
        }
      };

      warnAboutDataLoss = true;
    }

    public int showDialog() {
      status = JFileChooser.APPROVE_OPTION;
      parserPanel.resetTranslationTable();
      setVisible(true);
      return status;
    }

  }

  public TreeIODialog(Frame owner, boolean readOrWrite, boolean encapsulateXmlTagPanel) {
    parserPanel = new TreeXMLParserPanel(readOrWrite);
    setReadOrWrite(readOrWrite);
    newickIO = new NewickFormat();
    nexusIO  = new NexusFormat();
    xmlIO    = new XMLTreeFormat();
    pagelIO  = new PagelFormat();
    if(encapsulateXmlTagPanel) {
      parserDialog = new ParserDialog(owner, parserPanel);
    }
  }

  public void setReadOrWrite(boolean readOrWrite) {
    this.readOrWrite = readOrWrite;
    parserPanel.setReadOrWrite(this.readOrWrite);
    if(readOrWrite) {
      if(!Arrays.asList(super.getActionListeners()).contains(writeListener)) super.addActionListener(writeListener);
    } else {
      if(Arrays.asList(super.getActionListeners()).contains(writeListener)) super.removeActionListener(writeListener);
    }
  }

  public Tree[] readTree() throws Exception {
    String format = super.getSelectedFormat();
    File input = getSelectedFile();
    Tree[] trees = null;
    ITree iTree;
    if(format.equals("Newick"))  iTree = newickIO;
    else if(format.equals("Nexus" ))  iTree = nexusIO;
    else if(format.equals("XML"   ))  {
      if(parserDialog != null) if(parserDialog.showDialog() == JFileChooser.CANCEL_OPTION) return null; //Operation cancelled
      iTree = xmlIO;
      xmlIO.setTagDescription(parserPanel.getTagDescription());
    } else iTree = null; // this can't happen.
    trees = iTree.readTree(input);
    return trees;
  }

  public boolean writeTree(Tree tree) throws Exception {
    String format = super.getSelectedFormat();
    File output = getSelectedFile();
    return writeTree(tree, output, format);
  }
  public void writeTree(Tree[] trees) throws Exception {
    String format = super.getSelectedFormat();
    File output = getSelectedFile();
    this.writeTree(trees, output, format);
  }

  public boolean writeTree(Tree tree, File output) throws Exception {
    String format = super.getSelectedFormat();
    return writeTree(tree, output, format);
  }
  public void writeTree(Tree[] trees, File output) throws Exception {
    String format = super.getSelectedFormat();
    this.writeTree(trees, output, format);
  }

  public boolean writeTree(Tree tree, String format) throws Exception {
    File output = getSelectedFile();
    return writeTree(tree, output, format);
  }
  public void writeTree(Tree[] trees, String format) throws Exception {
    File output = getSelectedFile();
    this.writeTree(trees, output, format);
  }

  public boolean writeTree(Tree tree, File output, String format) throws Exception {
    if(!format.equals("XML")
     && XTree.class.isAssignableFrom(tree.getClass())
     && warnAboutDataLoss
     && XTreeTools.treeHasNonStandardAttributes((XTree)tree)
    ) {
      int response = JOptionPane.showConfirmDialog(
        this,
        "The tree contains non standard attributes that will not be saved in the selected format.\n Do you want to proceed?",
        "WARNING!!!",
        JOptionPane.YES_NO_OPTION,
        JOptionPane.QUESTION_MESSAGE
      );
      if(response == JOptionPane.NO_OPTION) {
        return false;
      }
    }
    if(format.equals("Newick")) {
      Tree treeToWrite = tree;
      if(!TreeTools.checkLeavesNames(tree, "()")) {
        int response = JOptionPane.showConfirmDialog(
          this,
          "Leave names must not contain parentheses when exporting to newick.\n Do you want to remove them in file?",
          "WARNING!!!",
          JOptionPane.YES_NO_OPTION,
          JOptionPane.QUESTION_MESSAGE
        );
        if(response == JOptionPane.YES_OPTION) {
          treeToWrite = (Tree)tree.clone();
          TreeTools.correctNames(treeToWrite, "()");
        }
      }
      newickIO.writeTree(treeToWrite, output);
    } else if(format.equals("XML")) {
      try {
        parserPanel.setAvailableTags(XTreeTools.getAvailableTags( (XTree) tree, false));
      } catch(ClassCastException cce) {
        // This is not a XTree object, ignore command.
      }
      if(parserDialog != null) parserDialog.setVisible(true);
      xmlIO.setTagDescription(parserPanel.getTagDescription());
      xmlIO.writeTree(tree, output);
    } else if(format.equals("Pagel")) {
      try {
        parserPanel.setAvailableTags(XTreeTools.getAvailableTags( (XTree) tree, false));
        if(parserDialog != null) parserDialog.setVisible(true);
        pagelIO.setTagDescription(parserPanel.getTagDescription());
        pagelIO.writeTree((XTree)tree, output);
      } catch(ClassCastException cce) {
        pagelIO.writeTree(tree, output);
      }
    }

    // Operation succeeded.
    return true;
  }
  public void writeTree(Tree[] trees, File output, String format) throws Exception {
    if(format.equals("Newick")) newickIO.writeTree(trees, output);
    if(format.equals("XML")) {
      xmlIO.setTagDescription(parserPanel.getTagDescription());
      xmlIO.writeTree(trees, output);
    } else if(format.equals("Pagel")) {
      pagelIO.setTagDescription(parserPanel.getTagDescription());
      pagelIO.writeTree(trees, output);
    }
  }

  public void setSelectedFormat(String format) throws IllegalArgumentException {
    selectedFormat = format;
    super.setSelectedFormat(format);
  }

  public int showOpenDialog(Component c) {
    setFormatsAvailable(READABLE_FORMATS_AVAILABLE, READABLE_FILTERS_AVAILABLE);
    if(selectedFormat != null) setSelectedFormat(selectedFormat);
    return super.showOpenDialog(c);
  }

  public int showSaveDialog(Component c) {
    setFormatsAvailable(WRITABLE_FORMATS_AVAILABLE, WRITABLE_FILTERS_AVAILABLE);
    if(selectedFormat != null) setSelectedFormat(selectedFormat);
    return super.showSaveDialog(c);
  }

  public ParserOptionPanel getParserOptionPanel() { return parserPanel; }

  private boolean readOrWrite;
  private NewickFormat newickIO;
  private NexusFormat nexusIO;
  private XMLTreeFormat xmlIO;
  private PagelFormat pagelIO;
  private TreeXMLParserPanel parserPanel;
  private ParserDialog parserDialog;
  private ActionListener writeListener;
  private String selectedFormat;
  private boolean warnAboutDataLoss;
  public static final String[] WRITABLE_FORMATS_AVAILABLE = { "Newick", "XML", "Pagel" };
  public static final String[] READABLE_FORMATS_AVAILABLE = { "Newick", "Nexus", "XML" };
  public static final FileFilter[] WRITABLE_FILTERS_AVAILABLE = { NewickFormat.NEWICK_FILTER, XMLTools.XML_FILTER, PagelFormat.PAGEL_FILTER };
  public static final FileFilter[] READABLE_FILTERS_AVAILABLE = { NewickFormat.NEWICK_FILTER, NexusFormat.NEXUS_FILTER, XMLTools.XML_FILTER };
}
