// -------------------------------------------------------------------------
//
// Copyright or © or Copr. CNRS, (November 16, 2004)
//
// This software is a library providing tool for bioinformatics.
//
// This software is governed by the CeCILL  license under French law and
// abiding by the rules of distribution of free software.  You can  use,
// modify and/ or redistribute the software under the terms of the CeCILL
// license as circulated by CEA, CNRS and INRIA at the following URL
// "http://www.cecill.info".
//
// As a counterpart to the access to the source code and  rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty  and the software's author,  the holder of the
// economic rights,  and the successive licensors  have only  limited
// liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading,  using,  modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean  that it is complicated to manipulate,  and  that  also
// therefore means  that it is reserved for developers  and  experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and,  more generally, to use and operate it in the
// same conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.
//
// -------------------------------------------------------------------------

package bio.dendogram;

/**
 * <p>Title: Dendogram</p>
 * <p>Description: Set of Tools for dealing with phylogenetic trees (dendograms).</p>
 * <p>Copyright: 2002</p>
 * <p>Company: </p>
 * @author Julien Dutheil
 * @version 1.0
 */

import gnu.jtools.stats.array.*;

public class GroupsManager {

  /**
   * A 'group' - sensu stricto - is a node with a name which is not a leaf.
   * The name of a leaf is the standard leaf name.
   * A 'group' - sensu lato - includes leaves: all the displayable nodes,
   * i.e. all the node except the root node.
   */

  public static boolean isGroup(Node node) { return node.hasName() && !node.isLeaf(); }

  public static int dimGroup(Node node) {
    int result = 0;
    if(isGroup(node) || node.isLeaf()) {//group
      result = 1;
    } else {//branche
      int[] d = new int[node.getNumberOfSons()];
      for(int i = 0; i < d.length; i++) {
        d[i] = dimGroup(node.getSon(i));
      }
      result = 1 + ArrayStatTools.max(d);
    }
    return result;
  }

  public static NodeList getGroups(Node node) {//Send the list of references to all groups.
    NodeList result = new NodeList();
    if(node.isLeaf()) return result; //No group.
    else {
      if(isGroup(node)) {//It's a group.
        result.add(node);
      }
      for(int i = 0; i < node.getNumberOfSons(); i++) result.addAll(getGroups(node.getSon(i)));
    }
    return result;
  }
  public static NodeList getGroups(Tree tree) {//Send the list of references to all groups.
    return getGroups(tree.getRootNode());
  }

  public static NodeList getLowerGroups(Node node) {//Send the list of references to lower groups.
    NodeList result = new NodeList();
    if(node.isLeaf()) return result;
    else {
      if(isGroup(node)) {
        result.add(node);
      } else {// Do not add internal groups.
        for(int i = 0; i < node.getNumberOfSons(); i++) result.addAll(getLowerGroups(node.getSon(i)));
      }
      return result;
    }
  }

  public static NodeList getHigherGroups(Node node) {//Send the list of references to higher groups.
    NodeList result = new NodeList();
    if(node.isLeaf()) return result;
    else {
      if(isGroup(node) && getGroups(node).size() == 1) {//Group with no internal group.
        result.add(node);
      } else {//Add internal groups.
        for(int i = 0; i < node.getNumberOfSons(); i++) result.addAll(getHigherGroups(node.getSon(i)));
      }
      return result;
    }
  }

  public static Node getGeneralSubtree(Node node) {//Groups are replaced by branches with the same name.
    Node result = (Node)node.clone();
    node.deleteFather();
    node.getSons().clear();
    if(!isGroup(node) && !node.isLeaf()) {//This is not a group...
      for(int i = 0; i < node.getNumberOfSons(); i++) Tree.addNode(node, getGeneralSubtree(node.getSon(i)));
    }
    return result;
  }
  public static Tree getGeneralTree(Tree tree) {//Groups are replaced by branches with the same name.
    Tree gt = (Tree)tree.clone();
    gt.setRootNode(getGeneralSubtree(tree.getRootNode()));
    return gt;
  }

  public static TaxonomicGroup listOfGroupsNames(Node node) {
    TaxonomicGroup result = new TaxonomicGroup(10);
    if(isGroup(node) || node.isLeaf()) {//une feuille
      if(node.hasName()) result.add(node.getName());
      //Sinon on laisse...
    } else {//une branche
      for(int i = 0; i < node.getNumberOfSons(); i++) {
        result.addAll(listOfGroupsNames(node.getSon(i)));
      }
    }
    return result;
  }
  public static TaxonomicGroup listOfGroupsNames(Tree tree) {
    return listOfGroupsNames(tree.getRootNode());
  }

  public static int getNumberOfGroups(Node node) {
    int result = 0;
    if(isGroup(node) || node.isLeaf()) {//group
      result = 1;
    } else {//branch
      for(int i = 0; i < node.getNumberOfSons(); i++) {
        result += getNumberOfGroups(node.getSon(i));
      }
    }
    return result;
  }
}
