// -------------------------------------------------------------------------
//
// Copyright or © or Copr. CNRS, (November 16, 2004)
//
// This software is a library providing tool for bioinformatics.
//
// This software is governed by the CeCILL  license under French law and
// abiding by the rules of distribution of free software.  You can  use,
// modify and/ or redistribute the software under the terms of the CeCILL
// license as circulated by CEA, CNRS and INRIA at the following URL
// "http://www.cecill.info".
//
// As a counterpart to the access to the source code and  rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty  and the software's author,  the holder of the
// economic rights,  and the successive licensors  have only  limited
// liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading,  using,  modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean  that it is complicated to manipulate,  and  that  also
// therefore means  that it is reserved for developers  and  experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and,  more generally, to use and operate it in the
// same conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.
//
// -------------------------------------------------------------------------

package bio.dendogram;

/**
 * <p>Title: Dendogram</p>
 * <p>Description: Set of Tools for dealing with phylogenetic trees (dendograms).</p>
 * <p>Copyright: 2002</p>
 * <p>Company: </p>
 * @author not attributable
 * @version 1.0
 */
import java.util.*;

public class Node {

  protected int id;
  protected String name;
  protected Double length, bootstrap;
  protected Node father;
  protected NodeList sons;
  protected int numberOfLeaves, dimension;

  public Node(int id) {
    this.id = id;
    name = null;
    length = null;
    bootstrap = null;
    sons = new NodeList(2);
    father = null;
    numberOfLeaves = 1;
    dimension = 0;
  }

  public Node(int id, String name) {
    this.id = id;
    this.name = name;
    length = null;
    bootstrap = null;
    sons = new NodeList(2);
    father = null;
    numberOfLeaves = 1;
    dimension = 0;
  }

  public Node(Node node) {
    id = node.id;
    if(node.hasName()) name = new String(node.getName());
    if(node.hasLength()) length = new Double(node.getLengthValue());
    if(node.hasBootstrap()) bootstrap = new Double(node.getBootstrapValue());
    if(node.sons != null) sons = new NodeList(node.sons); else sons = new NodeList(2);
    father = node.father;
    numberOfLeaves = node.numberOfLeaves;
    dimension = node.dimension;
  }

  public Object clone() {
    Node clone = new Node(id);
    if(name != null) clone.name = new String(name);
    if(length != null) clone.length = new Double(length.doubleValue());
    if(bootstrap != null) clone.bootstrap = new Double(bootstrap.doubleValue());
    if(sons != null) clone.sons = new NodeList(sons);
    clone.father = father;
    clone.numberOfLeaves = numberOfLeaves;
    clone.dimension = dimension;
    return clone;
  }

  public boolean equals(Node node) { return id == node.id; }

  /**
   * Get parameters:
   */

  public int getId() { return id; }
  public String getName() { return name; }
  public Double getLength() { return length; }
  public Double getBootstrap() { return bootstrap; }
  public double getLengthValue() throws NullPointerException { return getLength().doubleValue(); }
  public double getBootstrapValue() throws NullPointerException { return getBootstrap().doubleValue(); }
  public Node getFather() { return father; }
  public NodeList getSons() { return sons; }
  public Node getSon(int i) { return (Node)sons.get(i); }
  public int getNumberOfLeaves() { return numberOfLeaves; }
  public int getDimension() { return dimension; }
  public int getNumberOfSons() { return sons.size(); }

  /**
   * Set parameters:
   */

  public void setId(int id) { this.id = id; }
  public void setName(String name) { this.name = name; }
  public void setLength(double length) { this.length = new Double(length); }
  public void setBootstrap(double bootstrap) { this.bootstrap = new Double(bootstrap); }
  public void setFather(Node tree) { this.father = tree; }
  protected void setNumberOfLeaves(int numberOfLeaves) { this.numberOfLeaves = numberOfLeaves; }
  protected void setDimension(int dimension) { this.dimension = dimension; }

  /**
   * Delete parameters
   */

  public void deleteName()      { name = null; }
  public void deleteLength()    { length = null; }
  public void deleteBootstrap() { bootstrap = null; }
  public void deleteFather()    { father = null; }

  /**
   * Tests:
   */
  public boolean isLeaf()       { return getNumberOfSons() == 0; }
  public boolean hasName()      { return name      != null; }
  public boolean hasLength()    { return length    != null; }
  public boolean hasBootstrap() { return bootstrap != null; }
  public boolean hasFather()    { return father    != null; }

  /**
   * Edition:
   */
  public Node removeSon(int i) {
    Node node = (Node)sons.remove(i);
    node.deleteFather();
    return node;
  }
  public boolean removeSon(Node node) {
    boolean t = sons.remove(node);
    node.deleteFather();
    return t;
  }
  public void removeSons() { sons.clear(); }
  public void addSon(Node node) {
    sons.add(node);
    node.setFather(this);
  }
  public void addSon(int i, Node node) {
    sons.add(i, node);
    node.setFather(this);
  }

  /**
   * Display:
   */

  public void swap(int branch1, int branch2) {
    Node node1 = getSon(branch1);
    Node node2 = getSon(branch2);
    removeSon(node1);
    removeSon(node2);
    addSon(branch1, node2);
    addSon(branch2, node1);
  }
}
