// -------------------------------------------------------------------------
//
// Copyright or © or Copr. CNRS, (November 16, 2004)
//
// This software is a library providing tool for bioinformatics.
//
// This software is governed by the CeCILL  license under French law and
// abiding by the rules of distribution of free software.  You can  use,
// modify and/ or redistribute the software under the terms of the CeCILL
// license as circulated by CEA, CNRS and INRIA at the following URL
// "http://www.cecill.info".
//
// As a counterpart to the access to the source code and  rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty  and the software's author,  the holder of the
// economic rights,  and the successive licensors  have only  limited
// liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading,  using,  modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean  that it is complicated to manipulate,  and  that  also
// therefore means  that it is reserved for developers  and  experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and,  more generally, to use and operate it in the
// same conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.
//
// -------------------------------------------------------------------------

package bio.dendogram;

/**
 * <p>Title: Dendogram</p>
 * <p>Description: Set of Tools for dealing with phylogenetic trees (dendograms).</p>
 * <p>Copyright: 2002</p>
 * @author Julien Dutheil
 * @version 1.0
 */

/**
 * This Tree class comes from the original TreeEditor package, till the Baobab 1.41 version.
 * I have improved it a bit, so that in later version of Baobab, the Tree class used will
 * be another Tree class that will inherit this class.
 * This class is used in all the programs that need dealing with dendograms,
 * and not especially for editing purpose.
 */

import gnu.jtools.utils.file.*;
import gnu.jtools.stats.array.*;
import bio.sequences.*;
import bio.taxonomy.*;
import java.awt.event.*;
import java.awt.Color;
import java.util.*;
import java.io.*;

/**
 * A FEW NOTES:
 *
 * This class provides a few methods to deal with nodes (see class Node).
 * A tree as a pointer to a root node.
 *
 * We call hereafter a subtree a node and all its descendants.
 */


public class Tree {

  /**
   * Constructors:
   */

  /**
   * Construct a new void Tree object.
   */
  public Tree() {
    nextId = 0;
    name = null;
    root = null;
    modificationListeners = new ArrayList();
  }

  /**
   * Construct a new void Tree object of name 'name'.
   *
   * @param name the name of the new tree
   */
  public Tree(String name) {
    nextId = 0;
    this.name = name;
    root = null;
    modificationListeners = new ArrayList();
  }

  /**
   * Construct a new Tree object with root node 'root'.
   *
   * @param root the root node of the tree
   */
  public Tree(Node root) {
    nextId = 0;
    name = null;
    this.root = root;
    modificationListeners = new ArrayList();
  }

  /**
   * construct a new Tree object of name 'name' and root node 'root'.
   *
   * @param name the of the new tree
   * @param root the root node of the tree
   */
  public Tree(String name, Node root) {
    nextId = 0;
    this.name = name;
    this.root = root;
    modificationListeners = new ArrayList();
  }

  /**
   * Copy constructor.
   *
   * Construct a new Tree object which is a clone of 'tree'.
   *
   * @param tree the tree that will be cloned
   */
  public Tree(Tree tree) {
    this.root = convertSubtree(tree.getRootNode());
    if(tree.name!= null) name = new String(tree.name);
    modificationListeners = new ArrayList(tree.modificationListeners);
  }


  /**
   * Convert a node to subtree.
   *
   * All son nodes are cloned.
   * do the same a cloneSubtree, but does the father of the returned node is set to null.
   *
   * @param node the node defining the subtree to be cloned
   * @return a node defining the cloned subtree.
   * @see cloneSubtree
   */
  public final Node convertSubtree(Node node) {
    Node clone = createNode(node);
    clone.deleteFather();
    clone.removeSons();
    for (int i = 0; i < node.getNumberOfSons(); i++) {
      addNode(clone, convertSubtree(node.getSon(i)));
    }
    return clone;
  }

  /**
   * Clone a subtree.
   *
   * @param node the node defining the subtree to be cloned
   * @return a node defining the cloned subtree.
   * @see convertSubtree
   */
  public static final Node cloneSubtree(Node node) {
    Node clone = (Node)node.clone();
    clone.deleteFather();
    clone.removeSons();
    for (int i = 0; i < node.getNumberOfSons(); i++) {
      addNode(clone, cloneSubtree(node.getSon(i)));
    }
    return clone;
  }

  /**
   * Create a new node for this tree.
   *
   * Increment the node id counter.
   *
   * @return a new node with specified id
   */
  public Node createNode() { return new Node(nextId++); }

  /**
   * Create a new node with names 'name' for this tree.
   *
   * If checkName is set to true look if the name is already used.
   * Increment the node id counter.
   *
   * @param nodeName  The name of the node
   * @param checkName Tell if we have to check if the name is already used.
   * @return a new node with specified id
   * @throws NodeException If checkName is set to true and nodeName is already used in the tree.
   */
  public Node createNode(String nodeName, boolean checkName) throws NodeException {
    if(checkName) {
      if(TreeTools.subtreeContainsNodeName(root, nodeName)) throw new NodeException("Name already used: " + nodeName);
    }
    return new Node(nextId++, nodeName);
  }

  /**
   * Create a new node which is a clone of 'node'.
   *
   * Increment the node id counter.
   *
   * @param node the node to be cloned
   * @return a new node with specified id
   */
  public Node createNode(Node node) {
    //System.out.println("cloning node");
    Node newNode = new Node(node);
    newNode.setId(nextId++);
    return newNode;
  }

  /**
   * Get the root nod of this tree.
   *
   * @return the root node for this tree
   */
  public Node getRootNode() { return root; }

  /**
   * Set the root node of this tree.
   *
   * @param root the new root node
   */
  public void setRootNode(Node root) { this.root = root; }

  public String getName() { return name; }
  public void setName(String name) { this.name = name; }

  public Object clone() {
    //WARNING!!! This method clones all internal nodes too.
    Tree result = new Tree(this.cloneSubtree(root));
    result.name = name;
    return result;
  }

  /**
   * Collection-like methods:
   */

  public static boolean contains(Node subtree, Node node) {
    if(node == subtree) return true;
    else {
      for(int i = 0; i < subtree.getNumberOfSons(); i++) if(contains(subtree.getSon(i), node)) return true;
    }
    return false;
  }

  public boolean contains(Node node) { return contains(root, node); }


/*  public boolean equals(Object o) {
    /**
     * Two trees are equals if they contain the same branches.
     */
/*    if(o == null) return false;
    boolean result = true;
    Tree tree;
    try {
      tree = (Tree)o;
      if(tree.isLeaf()) {//tree is a leaf
        if(!tree.hasName()) return !this.hasName();
        result = this.isLeaf() && tree.getName().equals(this.getName()) ;
      } else {//tree is a subtree
        result = result && this.containsAll(tree);
      }
    } catch(ClassCastException e) {
      result = false; //'o' is not a Tree!
    }
    return result;
    }*/

  /**
   * Parameters:
   */

  public static void setLengthsOfSubtree(Node parentNode, double commonValue) {
    parentNode.setLength(commonValue);
    if(!parentNode.isLeaf()) {
      for(int i = 0; i < parentNode.getNumberOfSons(); i++) {
        setLengthsOfSubtree(parentNode.getSon(i), commonValue);
      }
    }
  }

  public void setLengthsOfTree(double commonValue) {
    setLengthsOfSubtree(root, commonValue);
  }

  public void setLengthsOfTreeExceptRoot(double commonValue) {
    for(int i = 0; i < root.getNumberOfSons(); i++) {
      setLengthsOfSubtree(root.getSon(i), commonValue);
    }
  }

  public static void setBootstrapsOfSubtree(Node parentNode, double commonValue) {
    if(!parentNode.isLeaf()) {
      parentNode.setBootstrap(commonValue);
      for(int i = 0; i < parentNode.getNumberOfSons(); i++) {
        setBootstrapsOfSubtree(parentNode.getSon(i), commonValue);
      }
    }
  }

  public void setBootstrapsOfTree(double commonValue) {
    setBootstrapsOfSubtree(root, commonValue);
  }

  public void setBootstrapsOfTreeExceptRoot(double commonValue) {
    for(int i = 0; i < root.getNumberOfSons(); i++) {
      setBootstrapsOfSubtree(root.getSon(i), commonValue);
    }
  }

  public static void deleteInternalNodesNames(Node parentNode) {
    if(!parentNode.isLeaf()) {
      parentNode.deleteName();
      for(int i = 0; i < parentNode.getNumberOfSons(); i++) {
        deleteInternalNodesNames(parentNode.getSon(i));
      }
    }
  }

  public void deleteInternalNodesNames() {
    deleteInternalNodesNames(root);
  }

  public static void deleteLengthsOfSubtree(Node parentNode) {
    parentNode.deleteLength();
    for(int i = 0; i < parentNode.getNumberOfSons(); i++) {
      deleteLengthsOfSubtree(parentNode.getSon(i));
    }
  }

  public void deleteLengthsOfTree() {
    deleteLengthsOfSubtree(root);
  }

  public static void deleteBootstrapsOfSubtree(Node parentNode) {
    parentNode.deleteBootstrap();
    for(int i = 0; i < parentNode.getNumberOfSons(); i++) {
      deleteBootstrapsOfSubtree(parentNode.getSon(i));
    }
  }

  public void deleteBootstrapsOfTree() {
    deleteBootstrapsOfSubtree(root);
  }

  /**
   * Tests:
   */

  public boolean hasLengths() { return TreeTools.hasLengths(root); }

  public boolean hasLengthsIgnoreRoot() {
    boolean[] result = new boolean[root.getNumberOfSons()];
    for(int i = 0; i < result.length; i++) {
      result[i] = TreeTools.hasLengths(root.getSon(i));
    }
    return ArrayOperators.inter(result);
  }

  public boolean hasBootstraps() { return TreeTools.hasBootstraps(root); }

  public int getNumberOfLeaves() { return root.getNumberOfLeaves(); }

  public int getDimension() { return root.getDimension(); }

  /**
   * Computes:
   */

  /****************************************************************************/

/*  public static void computeNumberOfLeaves(Node node) {
    int n;
    if(node.isLeaf()) n = 1;
    else {
      n = 0;
      for(int i = 0; i < node.getNumberOfSons(); i++) {
        computeNumberOfLeaves(node.getSon(i));
        n += node.getSon(i).getNumberOfLeaves();
      }
    }
    node.setNumberOfLeaves(n);
  }*:

  public void computeNumberOfLeaves() { computeNumberOfLeaves(root); }

  public static void actualizeNumberOfLeaves(Node node) {
    int n;
    if(node.isLeaf()) n = 1;
    else {
      n = 0;
      for(int i = 0; i < node.getNumberOfSons(); i++) {
        n += node.getSon(i).getNumberOfLeaves();
      }
    }
    node.setNumberOfLeaves(n);
    if(node.hasFather()) actualizeNumberOfLeaves(node.getFather());
  }*/

  public void actualizeNumberOfLeaves() {
    TreeTools.actualizeNumberOfLeavesUp(root);
  }

/*  public static void computeDimension(Node node) {
    int d;
    if(node.isLeaf()) d =0;
    else {
      d = node.getSon(0).getDimension();
      for(int i = 1; i < node.getNumberOfSons(); i++) {
        computeDimension(node.getSon(i));
        if(node.getSon(i).getDimension() > d) d = node.getSon(i).getDimension();
      }
    }
    node.setDimension(d);
  }

  public void computeDimension() { computeDimension(root); }

  public static void actualizeDimension(Node node) {
    int d;
    if(node.isLeaf()) d =0;
    else {
      d = node.getSon(0).getDimension();
      for(int i = 1; i < node.getNumberOfSons(); i++) {
        if(node.getSon(i).getDimension() > d) d = node.getSon(i).getDimension();
      }
    }
    node.setDimension(d);
    if(node.hasFather()) actualizeDimension(node.getFather());
  }*/

  public void actualizeDimension() {
    TreeTools.actualizeDimensionUp(root);
  }

  public static int actualizeId(Node node, int count) {
    node.setId(count++);
    for(int i = 0; i < node.getNumberOfSons(); i++) {
      count = actualizeId(node.getSon(i), count);
    }
    return count;
  }

  /**
   * Actualize all id tags for this tree.
   *
   * @param count The initial id (will be assigned to the root node).
   */
  public void actualizeId(int count) {
    nextId = actualizeId(root, count);
  }

  /**
   * Getting subtrees:
   */

  //Getting a Node from a Path
  public Node getNode(TreePath path) {
    Node result = root;
    for(int i = 0; i < path.size(); i++) {
      result = result.getSon(Integer.parseInt(""+path.get(i)));
    }
    return result;
  }

  public static int indexOf(Node son) { return son.getFather().getSons().indexOf(son); }

  public NodeList getLeaves() { return TreeTools.getLeaves(root); }

  public TaxonomicGroup listOfLeavesNames() { return TreeTools.listOfLeavesNames(root); }

  /**
   * Listeners:
   */

  public void addModificationListener(ActionListener listener) {
    if(!modificationListeners.contains(listener)) modificationListeners.add(listener);
  }

  public List getModificationListeners() { return modificationListeners; }

  public void treeIsModified(String command) {
    for(int i = 0; i < modificationListeners.size(); i++) {
      ((ActionListener)modificationListeners.get(i)).actionPerformed(new ActionEvent(this, 0, command));
    }
  }

  /**
   * Editing tree:
   */

  /**
   * Adding branch(es)
   */

  public static void addNode(Node father, Node son) {
    father.getSons().add(son);
    son.setFather(father);
    TreeTools.actualizeNumberOfLeavesDown(father);
    TreeTools.actualizeDimensionDown(father);
  }

  public static void addNode(Node father, Node son, int position) {
    father.addSon(son);
    TreeTools.actualizeNumberOfLeavesDown(father);
    TreeTools.actualizeDimensionDown(father);
  }

  public static void addAllNodes(Node father, NodeList sons) {
    for(int i = 0; i < sons.size(); i++) father.addSon((Node)sons.get(i));
    TreeTools.actualizeNumberOfLeavesDown(father);
    TreeTools.actualizeDimensionDown(father);
  }

  public void addNode(Node node, TreePath newBranchPath) {
    //the last position in the path is the position of the new branch.
    int newBranchPosition = ((Integer)newBranchPath.get(newBranchPath.size()-1)).intValue();
    newBranchPath.remove(newBranchPath.size()-1);
    addNode(node, newBranchPath, newBranchPosition);
    //newBranchPath points at the mother branch, according to overdefined method addBranch(see next).
  }

  public void addNode(Node son, TreePath motherBranchPath, int newBranchPosition) {
    addNode(getNode(motherBranchPath), son, newBranchPosition);
  }

  /**
   * <p>Insert the node 'node' according to path 'path'.
   * The node will be added between the two last elements of path.
   * This method creates a new node at the place of insertion.
   * </p>
   * @param node the node to be inserted
   * @param path the path where to add the node
   */
  public void insertNode(Node node, TreePath path) {
    //node will be added between the two last
    //nodes specified by the path.
    Node     newNode = createNode();
    Node currentNode = getNode(path);
    if(path.size() > 0) {
      if(currentNode.hasLength()) {
        newNode.setLength(currentNode.getLengthValue()/2);
        currentNode.setLength(currentNode.getLengthValue()/2);
      }
      cutNode(new TreePath(path)); //Current branch is removed.
                                   //Path is NOT corrected because of the cloning.
      addNode(newNode, currentNode);
      addNode(newNode, node);
      addNode(newNode, path);
      //new node is now at the former place of currentBranch.
    } else {//path points to the root:
      currentNode = createNode();
      if(node.hasLength()) {
        currentNode.setLength(node.getLengthValue()/2);
        node       .setLength(node.getLengthValue()/2);
      }
      addNode(currentNode, node);
      addNode(currentNode, root);
      root = currentNode;
    }
  }

  /**
   * Insert the node 'node' to the node 'currentNode'.
   * @param currentNode the node where to insert
   * @param node the node to insert
   * @see insertNode(Node node, TreePath path)
   */
  public void insertNode(Node currentNode, Node node) {
    TreePath path = getTreePath(currentNode);
    insertNode(node, path);
  }

  /**
   * Deleting branch(es):
   */
/*  private void reset() {
    deleteFather();
    deleteSons();//Remove all elements.
    deleteBootstrap();
    deleteLength();
    deleteName();
    numberOfLeaves = 0;
  }*/

  public Node cutNode(Node father, int sonPosition) {
    Node removed = father.removeSon(sonPosition);
    TreeTools.actualizeNumberOfLeavesDown(father);
    TreeTools.actualizeDimensionDown(father);
    return removed;
  }

  public boolean cutNode(Node son) {
    Node father = son.getFather();
    boolean removed = father.removeSon(son);
    TreeTools.actualizeNumberOfLeavesDown(father);
    TreeTools.actualizeDimensionDown(father);
    return removed;
  }

  public Node cutNode(TreePath branchPath) {
    int branchPosition = ((Integer)branchPath.get(branchPath.size()-1)).intValue();
    branchPath.remove(branchPath.size()-1);
    return cutNode(branchPath, branchPosition);
  }

  public Node cutNode(TreePath motherBranchPath, int branchPosition) {
    return cutNode(getNode(motherBranchPath), branchPosition);
  }

  public void exchangeSubtrees(Node node1, Node byNode2) {
    //if(node1 == root) System.err.println("ERROR: in Tree.exchangeSubtrees(Node, Node), can't exchange root node (node1).");
    //if(byNode2 == root) System.err.println("ERROR: in Tree.exchangeSubtrees(Node, Node), can't exchange root node (node2).");
    Node father = node1.getFather();
    int position = father.getSons().indexOf(node1);
    if(position == -1) System.err.println("ERROR: in Tree.replaceNode(Node, Node), first node is not valid.");
    else {
      byNode2.setFather(father);
      node1.deleteFather();
      cutNode(father, position);
      addNode(father, byNode2, position);
    }
  }

  public void removeNode(Node node) {
    Node father = node.getFather();
    cutNode(node);//the branch is removed.
    if(father.getNumberOfSons() == 1) {//node with a single branch.
      Node brother = cutNode(father, 0);//removing of the former sister branch.
      //Correction: mother becomes sister.
      if(brother.hasLength() && father.hasLength()) {
        brother.setLength(brother.getLengthValue() + father.getLengthValue());
      } else if(father.hasLength()) {
        brother.setLength(father.getLengthValue());
      } //Else keep sister length.
      //brother take the place of father,
      //father is removed from the tree.
      if(father == root) {
        brother.deleteFather();
        root = brother;
        father.deleteFather();
      } else {
        int position = indexOf(father);
        addNode(father.getFather(), brother, position);
        cutNode(father);
      }
    }
  }
  public void removeNode(TreePath path) {
    removeNode(getNode(path));
  }


  public void moveNode(Node node, Node withThisOne) {
    removeNode(node);
    addNode(withThisOne.getFather(), node);
  }

  /**
   * NEW DISPLAY
   */

  //swap branch at a particular node:
  public void swapAtNode(TreePath path, int branch1, int branch2) {
    Node node = getNode(path);
    node.swap(branch1, branch2);
  }

  public void displayRoot(Node newRoot) {
    //Start (general case):
    //              |----N1.1.1-...
    //    |----N1.1-|
    //    |         |----N1.1.2-...
    //    |         .
    //    |         |----N1.1.n-...
    // N1-|(=root)
    //    |----N1.2-...
    //    .
    //    |----N1.n-...
    //
    // Reroot with N1.1:
    //      |----N1.1.1-...
    //      |
    //      |----N1.1.2-...
    //      .
    //      |----N1.1.n-...
    // N1.1-|(=root)
    //      |       |----N1.2-...
    //      |----N1-|
    //              .
    //              |----N1.n-...
    //
    //Hence:
    // - if outgroup < N1.1, reroot with outgroup's father first
    // - else:
    //   - remove N1.1 from N1
    //   - move N1 to N1.1
    //   - set N1.1 as root
    // Done!
    //So here's the code:
    if(newRoot.getFather() != root) displayRoot(newRoot.getFather());
    else {
      cutNode(newRoot);
      addNode(newRoot, root);
      root = newRoot;
      actualizeNumberOfLeaves();
      actualizeDimension();
    }
  }

  public void newOutGroup(Node outgroup) throws NodeException {
    //Start (general case):
    //              |----N1.1.1-...
    //    |----N1.1-|
    //    |         |----N1.1.2-...
    //    |         .
    //    |         |----N1.1.n-...
    // N1-|(=root)
    //    |----N1.2-...
    //    .
    //    |----N1.n-...
    //
    // Reroot with N1.1.1:
    //      |----N1.1.1-...
    // newN-|(=root)
    //      |                 |----N1.2-...
    //      |         |----N1-.
    //      |         |       |----N1.n-...
    //      |----N1.1-|
    //                |----N1.1.2-...
    //                .
    //                |----N1.1.n-...
    //
    //Hence:
    // - if outgroup < N1.1.1, reroot with outgroup's father first
    // - else:
    //   - cut N1.1.1 from N1.1 (and not remove! We'll need N1.1 later...)
    //   - remove N1.1 from N1 (and not cut! if N1 has only two sons, N1 becomes N1.2))
    //   - add N1 to N1.1.
    //   - create new root node
    //   - add N1.1 to new root
    //   - add N1.1.1 to new root
    //   - set new root as root
    // Done!
    //So here's the code:
    if(outgroup == root) return; //nothing to do
    Node N11 = outgroup.getFather();
    if(N11 != root && N11.getFather() != root) newOutGroup(N11);
    Node N111 = outgroup;
    cutNode(N111);
    removeNode(N11);
    Node N1 = root;//Must be here and not before, because the may have change while removing node N11.
    addNode(N11, N1);
    Node newRoot = createNode();
    addNode(newRoot, N11);
    addNode(newRoot, N111);
    // Branch lengths:
    if(N11.hasLength() && N11.hasLength() && N111.hasLength()) {
      N1.setLength(N11.getLengthValue() / 2);
      N11.setLength(N11.getLengthValue() / 2 + N111.getLengthValue() / 2);
      N111.setLength(N111.getLengthValue() / 2);
    } else if(N11.hasLength() || N11.hasLength() || N111.hasLength()) {
      if(!N1.hasLength()) N1.setLength(0);
      if(!N11.hasLength()) N11.setLength(0);
      if(!N111.hasLength()) N111.setLength(0);
      throw new NodeException("Branch length missing, set to 0.");
    }
    actualizeNumberOfLeaves();
    actualizeDimension();
    setRootNode(newRoot);
  }
  public void newOutGroup(TreePath outgroup) throws NodeException { newOutGroup(getNode(outgroup)); }

  /**
   * GETTING SUBTREES AND PATHES
   */

/*  public Tree getSubtreeEqualTo(Tree research) {
    Tree result = null;
    if(this.equals(research)) return this;
    else {
      for(int i = 0; i < numberOfSons() && result == null; i++) result = this.branchAt(i).getSubtreeEqualTo(research);
      }
    return result;
    }*/

  public static TreePath getTreePath(Node parent, Node descendant) throws NullPointerException {
    if(parent == descendant) return new TreePath();//WARNING: rather equals than == ... ???
    else {
      if(parent.isLeaf()) {//sous-arbre non trouv�
        throw new NullPointerException();
      } else {
        TreePath path = null;
        for(int i = 0; i < parent.getNumberOfSons(); i++) {
          try {
            path = getTreePath(parent.getSon(i), descendant);
            path.add(0, new Integer(i));
          } catch(NullPointerException e) {}
        }
        if(path == null) throw new NullPointerException(); //sous-arbre non trouv� dans ce sous-arbre.
        else return path;
      }
    }
  }
  public TreePath getTreePath(Node descendant) throws NullPointerException { return getTreePath(root, descendant); }

  /**
   * MONOPHYLIE / PARAPHYLIE
   */

  public boolean isMonophyletic(TaxonomicGroup tg) {
    boolean result = false;
    ArrayList leaves = listOfLeavesNames();
    if(leaves.containsAll(tg) && leaves.size() == tg.size()) result = true;
    else {
      for(int i = 0; i < root.getNumberOfSons() && result == false; i++) result = TreeTools.isMonophyletic(root.getSon(i), tg);
    }
    return result;
  }

  public boolean isParaphyletic(TaxonomicGroup tg) {
    return (this.listOfLeavesNames().containsAll(tg) && !isMonophyletic(tg)) ? true : false;
  }


  protected List modificationListeners;
  protected Node root;
  protected int nextId;
  protected String name;
}
