// -------------------------------------------------------------------------
//
// Copyright or © or Copr. CNRS, (November 16, 2004)
//
// This software is a library providing tool for bioinformatics.
//
// This software is governed by the CeCILL  license under French law and
// abiding by the rules of distribution of free software.  You can  use,
// modify and/ or redistribute the software under the terms of the CeCILL
// license as circulated by CEA, CNRS and INRIA at the following URL
// "http://www.cecill.info".
//
// As a counterpart to the access to the source code and  rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty  and the software's author,  the holder of the
// economic rights,  and the successive licensors  have only  limited
// liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading,  using,  modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean  that it is complicated to manipulate,  and  that  also
// therefore means  that it is reserved for developers  and  experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and,  more generally, to use and operate it in the
// same conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.
//
// -------------------------------------------------------------------------

package bio.dendogram.datation;

import bio.dendogram.xml.*;

public class DatationTools {

  /**
   * Create (or reset) a 'Date' attribute for each node of the tree and set its value to 0.
   * @param tree the XTree to initialize.
   */
  public static void initDates(XTree tree) {
    XTreeTools.setAttributeToAllSubtrees(tree, DATE_ATTRIBUTE_NAME, new Double(0), false, true);
  }

  /**
   * Create (or reset) a 'Date' attribute for each node of a node and set its value to 0.
   * @param node the XNode defining the subtree to initialize.
   */
  public static void initDates(XNode node) {
    XTreeTools.setAttributeToAllSubtrees(node, DATE_ATTRIBUTE_NAME, new Double(0), false, true);
  }

  /**
   * Use 'Date' attributes to compute branch lengths of a tree.
   * @param tree the XTree to use.
   */
  public static void computeBranchLengths(XTree tree) {
    computeBranchLengths((XNode)tree.getRootNode());
  }

  /**
   * Use 'Date' attributes to compute branch lengths of a subtree.
   * @param node the XNode to use.
   */
  public static void computeBranchLengths(XNode node) throws NumberFormatException {
    if(!node.isLeaf()) {
      double thisDate;
      try {
        thisDate = ( (Double) node.getAttribute(DATE_ATTRIBUTE_NAME)).
            doubleValue();
      } catch(ClassCastException cce) {
        thisDate = Double.parseDouble(node.getAttribute(DATE_ATTRIBUTE_NAME).toString());
      }
      for(int i = 0; i < node.getNumberOfSons(); i++) {
        //For each son node:
        XNode son = (XNode)node.getSon(i);
        double sonDate = Double.parseDouble(son.getAttribute(DATE_ATTRIBUTE_NAME).toString());
        son.setLength(sonDate - thisDate);
        computeBranchLengths(son);
      }
    }
  }

  /**
   * Does the tree has 'Date' parameters at each node?
   * @param tree the tree to check
   * @return a boolean
   */
  public static boolean hasDates(XTree tree) {
    return XTreeTools.hasAttributeForEachSubtree(tree, DATE_ATTRIBUTE_NAME);
  }

  /**
   * Does the subtree defined by 'node' has 'Date' parameters at each subnode?
   * @param node the subtree to check
   * @return a boolean
   */
  public static boolean hasDates(XNode node) {
    return XTreeTools.hasAttributeForEachSubtree(node, DATE_ATTRIBUTE_NAME);
  }

  /**
   * Check if the tree is chronological, ie if inner node have earlier dates.
   * @param tree the tree to check
   * @return a boolean
   */
  public static boolean isChronological(XTree tree) {
    return isChronological((XNode)tree.getRootNode());
  }

  /**
   * Check if the subtree is chronological, ie if inner node have earlier dates.
   * @param node the sutree to check
   * @return a boolean
   */
  public static boolean isChronological(XNode node) {
    if(node.isLeaf()) return true; else {
      double thisDate;
      try {
        thisDate = ( (Double) node.getAttribute(DATE_ATTRIBUTE_NAME)).
            doubleValue();
      } catch(ClassCastException cce) {
        thisDate = Double.parseDouble(node.getAttribute(DATE_ATTRIBUTE_NAME).toString());
      }
      for(int i = 0; i < node.getNumberOfSons(); i++) {
        //For each son node:
        XNode son = (XNode)node.getSon(i);
        double sonDate;
        try {
          sonDate = ( (Double) son.getAttribute(DATE_ATTRIBUTE_NAME)).
              doubleValue();
        } catch(ClassCastException cce) {
          sonDate = Double.parseDouble(son.getAttribute(DATE_ATTRIBUTE_NAME).toString());
        }
        if(sonDate < thisDate || !isChronological(son)) return false;
      }
      return true;
    }
  }

  public static String DATE_ATTRIBUTE_NAME = "Date";
}
