package bio.sequences;

/**
 * <p>Title: Bio</p>
 * <p>Description: Tools for bioinformatics</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Julien Dutheil
 * @version 1.0
 */
import gnu.jtools.utils.string.TextTools;
import java.util.*;
import bio.*;

/**
 * @brief The sequence object. This is a general purpose container, containing a ordered
 * list of states(= letters). The states that allowed to be present in the sequence are defined
 * by an alphabet object, which is passed to the sequence constructor by a pointer.
 *
 * For programming convenience, the states are stored as integers, but the translation toward
 * and from a char description is easily performed with the Alphabet classes.
 *
 * Sequence objets also contain a name attribute and putativly some comments.
 *
 * @see Alphabet
 */
public class Sequence
    extends SymbolList {

  /**
   * @brief The sequence name.
   */
  protected String _name;

  /**
   * @brief The sequence comments.
   */
  protected ArrayList _comments;

  /**
   * @brief Sequence's sense (true : 5' -> 3'  /  false : 3' -> 5')
   */
  protected boolean _sense;

  /**
   * @brief Direct constructor: build a Sequence object from a string
   * You can use it safely for RNA, DNA and protein sequences.
   *
   * It can be used with codon sequences too, the string will be cut into
   * parts of size 3. But for more complicated alphabets, you should use one
   * complete constructors.
   *
   * @param name     The sequence name.
   * @param sequence The whole sequence to be parsed as a string.
   * @param alpha    A pointer toward the alphabet to be used with this sequence.
   */
  public Sequence(String name, String sequence, Alphabet alpha) throws BadCharException {
    super(alpha);
    _name = new String(name);
    setContent(sequence);
    _comments = null;
  }

  /**
   * @brief Direct constructor: build a Sequence object from a string
   * You can use it safely for RNA, DNA and protein sequences.
   *
   * It can be used with codon sequences too, the string will be cut into
   * parts of size 3. But for more complicated alphabets, you should use one
   * complete constructors.
   *
   * @param name     The sequence name.
   * @param sequence The whole sequence to be parsed as a string.
   * @param comments Comments to add to the sequence.
   * @param alpha    A pointer toward the alphabet to be used with this sequence.
   */
  public Sequence(String name, String sequence, List comments, Alphabet alpha) throws BadCharException {
    super(alpha);
    _name = new String(name);
    _comments = (comments == null) ? null : new ArrayList(comments);
    setContent(sequence);
  }

  /**
   * @brief General purpose constructor, can be used with any alphabet.
   *
   * You should note that the sequence is stored as a vector of int.
   * Hence each string in the vector will be translated using the alphabet object.
   *
   * @param name     The sequence name.
   * @param sequence The sequence content.
   * @param alpha    A pointer toward the alphabet to be used with this sequence.
   */
  public Sequence(String name, String[] sequence, Alphabet alpha) throws BadCharException {
    super(sequence, alpha);
    _name = new String(name);
    _comments = null;
  }

  /**
   * @brief General purpose constructor, can be used with any alphabet.
   *
   * You should note that the sequence is stored as a vector of int.
   * Hence each string in the vector will be translated using the alphabet object.
   *
   * @param name     The sequence name.
   * @param sequence The sequence content.
   * @param comments Comments to add to the sequence.
   * @param alpha    A pointer toward the alphabet to be used with this sequence.
   */
  public Sequence(String name, String[] sequence, List comments, Alphabet alpha) throws BadCharException {
    super(sequence, alpha);
    _name = new String(name);
    _comments = (comments == null) ? null : new ArrayList(comments);
  }

  /**
   * @brief General purpose constructor, can be used with any alphabet.
   *
   * @param name     The sequence name.
   * @param sequence The sequence content.
   * @param alpha    A pointer toward the alphabet to be used with this sequence.
   */
  public Sequence(String name, Vint sequence, Alphabet alpha) throws BadIntException {
    super(sequence, alpha);
    _name = new String(name);
    _comments = null;
  }

  /**
   * @brief General purpose constructor, can be used with any alphabet.
   *
   * @param name     The sequence name.
   * @param sequence The sequence content.
   * @param comments Comments to add to the sequence.
   * @param alpha    A pointer toward the alphabet to be used with this sequence.
   */
  public Sequence(String name, Vint sequence, List comments, Alphabet alpha) throws BadIntException {
    super(sequence, alpha);
    _name = new String(name);
    _comments = (comments == null) ? null : new ArrayList(comments);
  }

  /**
   * @brief The Sequence copy constructor. This does not perform a hard copy of the alphabet object.
   */
  public Sequence(Sequence s) {
    super(s);
    _name = new String(s.getName());
    _comments = (ArrayList)s.getComments();
  }

  /**
  * @name The Cloneable interface
  *
  * @{
  */
  public Object clone() {
    return new Sequence(this);
  }

  /** @} */


  /**
   * @name Setting/getting the name of the sequence.
   *
   * @{
   */

  /**
   * @brief Get the name of this sequence.
   *
   * @return This sequence name.
   */
  public String getName() {
    return new String(_name);
  }

  /**
   * @brief Set the name of this sequence.
   *
   * @param name The new name of the sequence.
   */
  public void setName(String name) {
    _name = new String(name);
  }

  /** @} */

  /**
   * @name Setting/getting the comments associated to the sequence.
   *
   * @{
   */

  /**
   * @brief Get the comments associated to this sequence.
   *
   * @return The comments of the sequence.
   */
  public List getComments() {
    return _comments == null ? null : new ArrayList(_comments);
  }

  /**
   * @brief Set the comments associated to this sequence.
   *
   * @param comments The new comments of the sequence.
   */
  public void setComments(List comments) {
    _comments = new ArrayList(comments);
  }

  /** @} */

  /**
   * @name Setting/getting the sense of the sequence.
   *
   * @{
   */

  /**
   * @brief Get the sense of the sequence.
   *
   * @return True is the sequence is 'sense', false if it is 'antisense'.
   */
  public boolean getSense() {
    return _sense;
  }

  /**
   * @brief Set the sense of the sequence: 'true' if 'sense', 'false' if 'antisense'.
   *
   * @param sense The sense of the sequence.
   */
  public void setSense(boolean sense) {
    _sense = sense;
  }

  /** @} */

  /**
   * @name Adjusting the size of the sequence.
   *
   * @{
   */

  /**
   * @brief Set the whole content of the sequence.
   *
   * @param sequence The new content of the sequence.
   * @see The Sequence constructor for information about the way sequences are internaly stored.
   */
  public void setContent(String sequence) throws BadCharException {
    // Remove blanks in sequence
    _content = StringSequenceTools.codeSequence(TextTools.removeBlanks(sequence), _alphabet); //Warning, an exception may be casted here!
  }

  /**
   * @brief Set up the size of a sequence from the right side.
   *
   * All new characters are filled with gaps.
   * If the specified size is < to the sequence size, the sequence will be truncated.
   *
   * @param size The new size of the sequence.
   */
  public void setToSizeR(int size) {
    int seqSize = _content.size();
    // Size verification
    if (size < seqSize) {
      _content.resize(size);
      return;
    }
    if (size == seqSize) return;

    // Add gaps up to specified size
    while(_content.size() < seqSize) _content.add(-1);
  }

  /**
   * @brief Set up the size of a sequence from the left side.
   *
   * All new characters are filled with gaps.
   * If the specified size is < to the sequence size, the sequence will be truncated.
   *
   * @param size The new size of the sequence.
   */
  public void setToSizeL(int size) {
    // Size verification
    int seqSize = _content.size();
    if (size < seqSize) {
      //We must truncate sequence from the left.
      //This is a very unefficient method!
      _content.erase(0, seqSize - size);
      return;
    }
    if (size == seqSize) return;

    // Add gaps up to specified size
    _content.insert(0, size - seqSize, -1);
  }

  /**
   * @brief Append the specified content to the sequence.
   *
   * @param content The content to append to the sequence.
   * @throw BadIntException If the content does not match the current alphabet.
   */
  void append(Vint content) throws BadIntException {
    // Check list for incorrect characters
    for(int i = 0; i < content.size(); i++) {
      if(!_alphabet.isIntInAlphabet(content.get(i))) throw new BadIntException(content.get(i), "Sequence::append", _alphabet);
    }
    //Sequence is valid:
    for(int i = 0; i < content.size(); i++) {
      _content.add(content.get(i));
    }
  }

  /**
   * @brief Append the specified content to the sequence.
   *
   * @param content The content to append to the sequence.
   * @throw BadCharException If the content does not match the current alphabet.
   */
  void append(String[] content) throws BadCharException {
    // Check list for incorrect characters
    for(int i = 0; i < content.length; i++) {
      if(!_alphabet.isCharInAlphabet(content[i])) throw new BadCharException(content[i], "Sequence::append", _alphabet);
    }
    //Sequence is valid:
    for (int i = 0; i < content.length; i++) {
      _content.add(_alphabet.charToInt(content[i]));
    }
  }

  /**
   * @brief Append the specified content to the sequence.
   *
   * @param content The content to append to the sequence.
   * @throw BadCharException If the content does not match the current alphabet.
   */
  public void append(String content) throws BadCharException {
    try { append(StringSequenceTools.codeSequence(content, _alphabet)); }
    catch(BadIntException bie) { bie.printStackTrace(); } // Never thrown.
  }

  /** @} */
}
