package bio.sequences;

/**
 * <p>Title: Bio</p>
 * <p>Description: Tools for bioinformatics</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Julien Dutheil
 * @version 1.0
 */

import java.util.*;
import bio.*;

/**
 * @brief The SequenceContainer interface.
 *
 * Interface for sequence containers manipulation.
 * This interface is the most general one, it supposes no ordering of the sequences in the
 * container. Sequences are retrievable using their names, which must be unic.
 *
 * The container is the SOLE responsible for the allocation/destruction of sequences it
 * contains. This means that any sequence passed to it will be COPIED into the container.
 * The container also provides methods that send pointers towards these sequences
 * (without performing any copy of the underlying objects).
 *
 * Note :
 * <ol>
 * <li> methods for adding sequences to the container are not declared here
 * (so they can't be used throught this interface),
 * because these methods take sequence container's type specific parameters
 * (i.e. a key for map sequence containers);</li>
 * <li> although a sequence can be deleted directly from the sequence container,
 * using specific methods (removeSequence and deleteSequence) to do it is
 * IMPERATIVE (because they achieve	verifications, and they properly
 * delete pointer toward the corresponding sequences).
 * </li>
 * </ol>
 *
 * @see Sequence
 */

public interface SequenceContainer {

  /**
   * @brief Get sequence container's alphabet.
   *
   * @return The alphabet associated to this container.
   */
  public Alphabet getAlphabet();

  /**
  * @brief Get the content of a sequence.
  *
  * @param name The name of the sequence.
  * @return The content of the sequence as a vector of integers.
  * @throw SequenceNotFoundException If the name does not match any sequence in the container.
  */
 public Vint getContent(String name) throws SequenceNotFoundException;

 /**
  * @brief Convert a particular sequence to a string.
  *
  * @param name The name of the sequence.
  * @return A string describing the content of the sequence.
  * @throw SequenceNotFoundException If the name does not match any sequence in the container.
  */
 public String toString(String name) throws SequenceNotFoundException;

 /**
  * @brief Retrieve a sequence object from the container.
  *
  * @param name The name of the sequence.
  * @return A pointer toward the Sequence with corresponding name.
  * @throw SequenceNotFoundException If the name does not match any sequence in the container.
  */
 public Sequence getSequence(String name) throws SequenceNotFoundException;

 /**
  * @brief Replace a sequence in the container.
  *
  * @param name      The name of the sequence.
  * @param sequence  The sequence to add.
  * @param checkName Tell if the container must check if the name of the sequence
  * is already used in the container before adding it.
  * @throw SequenceNotFoundException If the name does not match any sequence in
  * the container.
  * @throw Exception Any other kind of exception, if the name of the sequence is
  * already used, are whatever else depending on the implementation.
  */
 public void setSequence(String name, Sequence sequence, boolean checkName) throws Exception;

 /**
  * @brief Extract (and remove) a sequence from the container.
  *
  * @param name The name of the sequence.
  * @throw SequenceNotFoundException If the name does not match any sequence in
  * the container.
  */
 public Sequence removeSequence(String name) throws SequenceNotFoundException;

 /**
  * @brief Delete a sequence of the container.
  *
  * @param name The name of the sequence.
  * @throw SequenceNotFoundException If the name does not match any sequence in
  * the container.
  */
 public void deleteSequence(String name) throws SequenceNotFoundException;

 /**
  * @brief Get the number of sequences in the container.
  *
  * @return The number of sequences in the container.
  */
 public int getNumberOfSequences();

 /**
  * @brief Get all the names of the sequences in the container.
  *
  * @return A vector of strings with all sequence names.
  */
 public String[] getSequencesNames();

 /**
  * @brief Set all sequence names.
  *
  * @param names A vector of strings with all sequence names.
  * Its size must be strictly equal to the the size of the container (the number of
  * sequences).
  * @param checkName Tell if the container must check if the name of the sequence
  * is already used in the container before adding it.
  * @throw Exception If there are redundant names in the input vector.
  */
 public void setSequencesNames(String[] names, boolean checkNames) throws Exception;

 /**
  * @brief Get comments of a particular sequence.
  *
  * @param name The name of the sequence.
  * @return The comments associated to sequence with name 'name'.
  * @throw SequenceNotFoundException If the name does not match any sequence in
  * the container.
  */
 public List getComments(String name) throws SequenceNotFoundException;

 /**
  * @brief Set the comments of a particular sequence.
  *
  * @param name     The name of the sequence.
  * @param comments The comments to set to sequence with name 'name'.
  * @throw SequenceNotFoundException If the name does not match any sequence in
  * the container.
  */
 public void setComments(String name, List comments) throws SequenceNotFoundException;

 /**
  * @brief Get the comments of this container.
  *
  * @return The comments associated to this container.
  */
 public List getGeneralComments();

 /**
  * @brief Set the comments of this container.
  *
  * @param comments The comments to be associated to this container.
  */
 public void setGeneralComments(List comments);

 /**
  * @brief Delete the comments associated to this container.
  */
 public void deleteGeneralComments();

 /**
  * @brief Delete all sequences in the container.
  */
 public void clear();
};
