package bio.sequences;

import gnu.jtools.utils.string.TextTools;
import bio.*;

public class StringSequenceTools {

  /**
   * @brief Get a subsequence.
   *
   * @param sequence The input sequence.
   * @param begin    The begining position (included).
   * @param end      The ending position (included).
   * @return A string with the subsequence.
   * @throw Exception If position does not not match the interval [0, length].
   */
  public static String subseq(String sequence, int begin, int end) throws Exception {
    // Checking interval
    if (end < begin) throw new Exception("StringSequenceTools::subseq: Invalid interval");
    return sequence.substring(begin, end + 1);
  }

  /**
   * @brief Set up the size of a sequence from the right side.
   *
   * All new characters are filled with gaps.
   * If the specified size is < to the sequence size, the sequence will be truncated.
   *
   * @param size The new size of the sequence.
   */
  public static String setToSizeR(String sequence, int size) {
    return TextTools.resizeRight(sequence, size, '-');
  }

/**
   * @brief Set up the size of a sequence from the left side.
   *
   * All new characters are filled with gaps.
   * If the specified size is < to the sequence size, the sequence will be truncated.
   *
   * @param size The new size of the sequence.
   */
  public static String setToSizeL(String sequence, int size) {
    return TextTools.resizeLeft(sequence, size, '-');
  }

  /**
   * @brief Delete all occurence of a character in the sequence.
   *
   * @param sequence The sequence to parse.
   * @param chars    The character to remove.
   * @return         The sequence with all specified characters removed.
   */
  public static String deleteChar(String sequence, char chars) {
    // Copy sequence
    StringBuffer result = new StringBuffer(sequence);

    // Search and delete specified char
    for (int i = 0 ; i < result.length() ; i++) { if (result.charAt(i) == chars) result.deleteCharAt(i); }

    return result.toString();
  }

  /**
   * @brief Delete all occurence of several characters in the sequence.
   *
   * @param sequence The sequence to parse.
   * @param chars    The characters to remove.
   * @return         The sequence with all specified characters removed.
   */
  public static String deleteChar(String sequence, String chars) {
    // Copy sequence
    StringBuffer result = new StringBuffer(sequence);

    // For all characters to delete
    for (int i = 0 ; i < chars.length() ; i++) {
      // Search and delete char
      for (int j = 0 ; j < result.length() ; j++) { if (result.charAt(j) == chars.charAt(i)) result.deleteCharAt(j); }
    }

    return result.toString();
  }

  /**
   * @brief Reverse the sequence.
   *
   * @param sequence The sequence to reverse.
   * @return The reversed sequence.
   */
  public static String reverse(String sequence) {
    // Initializing
    String result = new String();

    // Main loop : reverse all characters of sequence
    int size = sequence.length();
    for (int i = 0 ; i < size ; i++) { result += sequence.charAt(size - i - 1); }

    // Send result
    return result;
  }

// !!! Upgrade to Translators! !!!
  /**
   * @brief Get the complement of a sequence.
   *
   * For this method, sequence is supposed to be of type DNA.
   *
   * @param sequence The sequence to complement.
   * @return The complementary sequence.
   */
  public static String complement(String sequence) {
    // Initializing
    String result = new String();

    // Main loop : completement all characters
    int size = sequence.length();
    for (int i = 0 ; i < size ; i++) {
      switch (sequence.charAt(i)) {
        case 'A' : result += 'T';
          break;
        case 'C' : result += 'G';
          break;
        case 'G' : result += 'C';
          break;
        case 'T' : result += 'A';
          break;
        case 'M' : result += 'K';
          break;
        case 'R' : result += 'Y';
          break;
        case 'Y' : result += 'R';
          break;
        case 'K' : result += 'M';
          break;
        case 'V' : result += 'B';
          break;
        case 'H' : result += 'D';
          break;
        case 'D' : result += 'H';
          break;
        case 'B' : result += 'V';
          break;
        default  : result += sequence.charAt(i);
          break;
      }
    }

    // Send new sequence
    return result;
  }

  /**
   * @brief Convert a string sequence to a vector of int.
   *
   * This method is used in the sequence constructor.
   * This deals with the most simple cases: the sequence (string) contains
   * a succession of all characters. This is indeed the case for simple
   * alphabets, but may be more complicated if the alphabet is coded with
   * variable code length.
   *
   * @param sequence The sequence to parse.
   * @param alphabet The alphabet to use to code the sequence.
   * @return A vector of int codes.
   * @throw BarCharException If some character does not match the specified alphabet.
   */
  public static Vint codeSequence(String sequence, Alphabet alphabet) throws BadCharException {
    try {
      int size = AlphabetTools.getAlphabetCodingSize(alphabet); //Warning, an exception may be casted here!
      Vint code = new Vint();
      int pos = 0;
      while (pos + size <= sequence.length()) {
        code.add(alphabet.charToInt(sequence.substring(pos, pos + size)));
        pos += size;
      }
      return code;
    } catch(Exception e) {
      e.printStackTrace();
    }
    return null;
  }

  /**
   * @brief Convert a sequence to its string representation.
   *
   * @param sequence The sequence object to convert.
   * @param alphabet The alphabet to use to decode the sequence.
   * @return A string representation of the sequence.
   * @throw BarIntException If some value does not match the specified alphabet.
   */
  public static String decodeSequence(Vint sequence, Alphabet alphabet) throws BadIntException {
    String result = "";
    for(int i = 0; i < sequence.size(); i++) {
      result += alphabet.intToChar(sequence.get(i));
    }
    return result;
  }

  /**
   * @brief Parse a sequence and try to guess the correct alphabet to use.
   *
   * @param sequence The sequence to parse.
   * @return A pointer toward a new Alphabet object.
   * @throw EmptySequenceException if the sequence is empty.
   * @throw SequenceException if their is an ambiguity between several alphabet.
   * @throw AlphabetException if the sequence does not match any alphabet.
   */
  public static Alphabet getAlphabetFromSequence(String sequence) throws EmptySequenceException, SequenceException, AlphabetException {
    // empty sequence test
    if (sequence.length() == 0) {
      throw new EmptySequenceException("Sequence::getAlphabetFromSequence : Empty sequence string", null);
    }

    // initialisation
    int nucleic = 0;	  // number of nucleic characters
    int proteic = 0;	  // number of proteic characters
    int unresolved = 0;   // number of unresolved characters (can be nucleic or proteic)
    boolean tLetter = false; // boolean indicates that a T letter is found
    boolean uLetter = false; // boolean indicates that the U nucleotid is found
    boolean unknow = false;  // boolean indicates that a unknow character is found

    // Main loop : for all character in sequence
    for (int i = 0 ; i < sequence.length() ; i++) {
      // Character analyse
      switch (AlphabetTools.getType(sequence.charAt(i))) {
        // if character can't be nucleic or proteic
        case -1 : unknow = true;
          break;
          // if character can't identify identify alphabet type (letters A, C, D, G, H, K, M, N, R, S, V, W, X, Y, -)
        case 0 : unresolved++;
          break;
          // if character identify nucleic alphabet (letters B, O, ?, 0)
        case 1 : nucleic++;
          break;
          // if character may identify DNA alphabet (letter T)
        case 2 : tLetter = true;
          unresolved++;
          break;
          // if character identify RNA alphabet (letter U)
        case 3 : uLetter = true;
          nucleic++;
          break;
          // if character identify proteic alphabet (letters E, F, I, L, P, Q)
        case 4 : proteic++;
          break;
      }
    }

    if (unknow) {
      throw new AlphabetException ("Sequence::getAlphabetFromSequence : Unknow character detected in specified sequence");
    }

    if ((nucleic > 0) && (proteic == 0)) {
      if (tLetter && !uLetter) return new DNA();
      if (!tLetter && uLetter) return new RNA();
      if (tLetter && uLetter) {
        throw new SequenceException ("Sequence::getAlphabetFromSequence : Confused sequence types found", null);
      }

      // if (!tLetter && !uLetter) : heuristic method :
      // if more than 95% of sequence is nucleic or unresolved, sequence is probably DNA
      if (((nucleic + unresolved) / sequence.length()) > 0.95) return (new DNA());
    }

    if ((nucleic == 0) && (proteic > 0)) {
      if (uLetter) {
        throw new SequenceException ("getAlphabetFromSequence : Confused sequence types found", null);
      }
      if (tLetter) return new ProteicAlphabet();

      // if (!tLetter && !uLetter) : heuristic method :
      // if more than 95% of sequence if proteic or unresolved, sequence if probably proteic
      if (((proteic + unresolved) / sequence.length()) > 0.95) return new ProteicAlphabet();
    }

    if ((nucleic == 0) && (proteic == 0) && tLetter) return new DNA();

    throw new SequenceException ("Sequence::getAlphabetFromSequence : Unresolved sequence type", null);
  }
}
