package bio.sequences;

/**
 * <p>Title: Bio</p>
 * <p>Description: Tools for bioinformatics</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Julien Dutheil
 * @version 1.0
 */

import bio.*;

/**
 * @brief The SymbolList object. This is a general purpose container, containing a ordered
 * list of states(= letters). The states that allowed to be present in the list are defined
 * by an alphabet object, which is passed to the list constructor by a pointer.
 *
 * For programming convenience, the states are stored as integers, but the translation toward
 * and from a char description is easily performed with the Alphabet classes.
 *
 * @see Alphabet
 */
class SymbolList
    implements Cloneable {

  /**
   * @brief The Alphabet attribute must be initialized in constructor and then never can't be changed.
   * (to apply another alphabet to a list you'll have to create a new list).
   */
  protected Alphabet _alphabet;

  /**
   * @brief The list content.
   */
  protected Vint _content;

  /**
   * @brief Build a new void SymbolList object with the specified alphabet.
   *
   * @param alpha The alphabet to use.
   */
  public SymbolList(Alphabet alpha) {
    _alphabet = alpha;
  }

  /**
   * @brief Build a new SymbolList object with the specified alphabet.
   * The content of the site is initialized from a vector of characters.
   *
   * @param list     The content of the site.
   * @param alpha    The alphabet to use.
   * @throw BadCharException If the content does not match the specified alphabet.
   */
  public SymbolList(String[] list, Alphabet alpha) throws BadCharException {
    _alphabet = alpha;
    setContent(list);
  }

  /**
   * @brief Build a new SymbolList object with the specified alphabet.
   * The content of the site is initialized from a vector of integers.
   *
   * @param list     The content of the site.
   * @param alpha    The alphabet to use.
   * @throw BadIntException If the content does not match the specified alphabet.
   */
  public SymbolList(Vint list, Alphabet alpha) throws BadIntException {
    _alphabet = alpha;
    setContent(list);
  }

  /**
   * @brief The copy constructor.
   */
  public SymbolList(SymbolList list) {
    _content  = list.getContent();
    _alphabet = list.getAlphabet();
  }

  /**
   * @name The Cloneable interface
   *
   * @{
   */
  public Object clone() {
    return new SymbolList(this);
  }

  /** @} */

  /**
   * @brief Get the alphabet associated to the list.
   *
   * @return A const pointer to the alphabet.
   * @see Alphabet class.
   */
  public Alphabet getAlphabet() {
    return _alphabet;
  }

  /**
   * @brief Get the number of elements in the list.
   *
   * @return The number of sites in the list.
   */
  public int size() {
    return _content.size();
  }

  /**
   * @name Acting on the content of the list.
   *
   * @{
   */

  /**
   * @brief Get the whole content of the list as a vector of int.
   *
   * @return list content.
   */
  public Vint getContent() {
    return new Vint(_content);
  }

  /**
   * @brief Set the whole content of the list.
   *
   * @param list The new content of the list.
   * @see The list constructor for information about the way lists are internaly stored.
   */
  public void setContent(Vint list) throws BadIntException {
    // Check list for incorrect characters
    for(int i = 0; i < list.size(); i++) {
      if(!_alphabet.isIntInAlphabet(list.get(i))) throw new BadIntException(list.get(i), "SymbolList::setContent", _alphabet);
    }
    //Sequence is valid:
    _content = new Vint(list);
  }

  /**
   * @brief Set the whole content of the list.
   *
   * @param list The new content of the list.
   * @see The list constructor for information about the way lists are internaly stored.
   */
  public void setContent(String[] list) throws BadCharException {
    // Check list for incorrect characters
    Vint coded = new Vint(list.length);
    for(int i = 0; i < list.length; i++) {
      if(!_alphabet.isCharInAlphabet(list[i])) throw new BadCharException(list[i], "SymbolList::setContent", _alphabet);
    }
    for(int i = 0; i < list.length; i++) {
      coded.set(i, _alphabet.charToInt(list[i]));
    }
    //SymbolList is valid:
    _content = coded;
  }

  /** @} */

  /**
   * @brief Convert the list as a string.
   *
   * This method is useful for dumping a list to a file or to the screen for display.
   *
   * @return The whole list as a string.
   */
  public String toString() {
    try {
      return StringSequenceTools.decodeSequence(_content, _alphabet);
    } catch(Exception e) {
      e.printStackTrace();
    }
    // Never reached!
    return null;
  }

  /**
   * @name Edition methods.
   *
   * @{
   */

  /**
   * @brief Add a character to the end of the list.
   *
   * @param c The character to add, given as a string.
   */
  public void addElement(String c) throws BadCharException {
    _content.add(_alphabet.charToInt(c));
  }

  /**
   * @brief Add a character at a certain position in the list.
   *
   * @param pos The postion where to insert the element.
   * @param c   The character to add, given as a string.
   */
  public void addElement(int pos, String c) throws BadCharException, IndexOutOfBoundsException {
    _content.add(pos, _alphabet.charToInt(c));
  }

  /**
   * @brief Set the element at position 'pos' to character 'c'.
   *
   * @param pos The position of the character to set.
   * @param c   The value of the element, given as a string.
   */
  public void setElement(int pos, String c) throws BadCharException, IndexOutOfBoundsException {
    _content.set(pos, _alphabet.charToInt(c));
  }

  /**
   * @brief Delete the element at postion 'pos'.
   *
   * @param pos The position of the element to delete.
   */
  public void deleteElement(int pos) throws IndexOutOfBoundsException {
     _content.remove(pos);
  }

  /**
   * @brief Get the element at position 'pos' as a character.
   *
   * @param pos The position of the character to retrieve.
   */
  public String getChar(int pos) throws IndexOutOfBoundsException {
    String c = "";
    try {
      c = _alphabet.intToChar(_content.get(pos));
    } catch(BadIntException bie) {
      //This should never happen!
    }
    return c;
  }

  /**
   * @brief Add a character to the end of the list.
   *
   * @param v The character to add, given as an int.
   */
  public void addElement(int v) throws BadIntException {
    //test:
    _alphabet.intToChar(v);
    _content.add(v);
  }

  /**
   * @brief Add a character at a certain position in the list.
   *
   * @param pos The postion where to insert the element.
   * @param v   The character to add, given as an int.
   */
  public void addElement(int pos, int v) throws BadIntException, IndexOutOfBoundsException {
    //test:
    _alphabet.intToChar(v);
    _content.add(pos, v);
  }

  /**
   * @brief Set the element at position 'pos' to character 'v'.
   *
   * @param pos The position of the character to set.
   * @param v   The value of the element, given as an int.
   */
  public void setElement(int pos, int v) throws BadIntException, IndexOutOfBoundsException {
    //test:
    _alphabet.intToChar(v);
    _content.set(pos, v);
  }

  /**
   * @brief Get the element at position 'pos' as an int.
   *
   * @param pos The position of the character to retrieve.
   */
  public int getValue(int pos) throws IndexOutOfBoundsException {
    return _content.get(pos);
  }

  /** @} */
}
