package bio.sequences;

/**
 * <p>Title: Bio</p>
 * <p>Description: Tools for bioinformatics</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Julien Dutheil
 * @version 1.0
 */

import java.util.*;
import bio.*;

/**
 * @brief The VectorSequenceContainer class.
 *
 * This is the simplest implementation of the OrderedSequenceContainer interface.
 * Sequences are stored in a vector of pointers.
 * The container is responsible for the creation and destruction of the sequence
 * objects it contains.
 */
public class VectorSequenceContainer
    extends AbstractSequenceContainer {

  /**
   * @brief A vector of pointers toward the sequences stored in the container.
   */
  protected ArrayList _sequences;

  /**
   * @brief Build a new container from a vector of pointers toward sequence objects.
   *
   * The addSequence() method is called uppon each Sequence object, hence each sequence is
   * <u>copied<u> into the container.
   *
   * @param vs    The list of sequence objects.
   * @param alpha The alphabet to all sequences.
   * @throw AlphabetMismatchException if one sequence does not match the specified alphabet.
   */
  public VectorSequenceContainer(List vs, Alphabet alpha) throws AlphabetMismatchException, ClassCastException, Exception {
    super(alpha);
    _sequences = new ArrayList();
    for(int i = 0; i < vs.size(); i++) addSequence((Sequence)vs.get(i), true);
  }

  /**
   * @brief Build a void container that will contain sequences of a particular alphabet.
   *
   * @param alpha The alphabet of the container.
   */
  public VectorSequenceContainer(Alphabet alpha) {
    super(alpha);
    _sequences = new ArrayList();
  }

  /**
   * @name Copy contructors:
   *
   * @{
   */

  /**
   * @brief Copy from a VectorSequenceContainer.
   *
   * @param vsc The VectorSequenceContainer to copy into this container.
   */
  public VectorSequenceContainer(VectorSequenceContainer vsc) {
    super(vsc.getAlphabet());
    int max = vsc.getNumberOfSequences();
    try { for(int i = 0 ; i < max ; i++) addSequence(vsc.getSequence(i), false); }
    catch(Exception e) { e.printStackTrace(); } // Never thrown
    _comments = new ArrayList(vsc.getGeneralComments());
  }

  /**
   * @brief Copy from an OrderedSequenceContainer.
   *
   * @param osc The OrderedSequenceContainer to copy into this container.
   */
  public VectorSequenceContainer(OrderedSequenceContainer osc) {
    super(osc.getAlphabet());
    // Sequences insertion
    int max = osc.getNumberOfSequences();
    try { for(int i = 0 ; i < max ; i++) addSequence(osc.getSequence(i), false); }
    catch(Exception e) { e.printStackTrace(); } // Never thrown

    _comments = new ArrayList(osc.getGeneralComments());
  }

  /** @} */

  /**
   * @name The SequenceContainer interface.
   *
   * @{
   */
  public Sequence getSequence(String name) throws SequenceNotFoundException {
    // Specified sequence name research into all sequences
    for(int i = 0; i < _sequences.size(); i++) {
      if(((Sequence)_sequences.get(i)).getName().equals(name)) return getSequence(i);
    }
    throw new SequenceNotFoundException("VectorSequenceContainer::getSequence : Specified sequence doesn't exist", name);
  }

  public void setSequence(String name, Sequence sequence, boolean checkNames) throws Exception {
    // Sequence's name existence checking
    int i = getSequencePosition(name);
    setSequence(i, sequence, checkNames);
  }

  public Sequence removeSequence(String name) throws SequenceNotFoundException {
    // Sequence's name existence checking
    int i = getSequencePosition(name);
    return removeSequence(i);
  }

  public void deleteSequence(String name) throws SequenceNotFoundException {
    // Sequence's name existence checking
    int i = getSequencePosition(name);
    deleteSequence(i);
  }

  public int getNumberOfSequences() { return _sequences.size(); }

  public String[] getSequencesNames() {
    String[] names = new String[getNumberOfSequences()];
    for(int i = 0; i < getNumberOfSequences(); i++) names[i] = getSequence(i).getName();
    return names;
  }

  public void setSequencesNames(String[] names, boolean checkNames) throws Exception {
    if(names.length != getNumberOfSequences()) throw new BadIntegerException("VectorSequenceContainer::setSequenceNames : bad number of names", names.length);
    if(checkNames) {
      for(int i = 0; i < names.length; i++) {
        // For all names in vector : throw exception if name already exists
        for(int j = 0; j < i; j++) {
          if(names[j] == names[i]) throw new Exception("VectorSiteContainer::setSequencesNames : Sequence's name already exists in container");
        }
      }
    }

    for(int i = 0; i < names.length; i++) {
      getSequence(i).setName(names[i]);
    }
  }

  public void clear() { _sequences.clear(); }

  /** @} */


  /**
   * @name The OrderedSequenceContainer interface.
   *
   * @{
   */
  public int getSequencePosition(String name) throws SequenceNotFoundException {
    // Specified sequence name research into all sequences
    for(int i = 0; i < _sequences.size(); i++) {
      if(getSequenceRef(i).getName().equals(name)) return i;
    }
    throw new SequenceNotFoundException("VectorSequenceContainer::getSequencePosition : Specified sequence doesn't exist", name);
  }

  public void setComments(int i, List comments) throws IndexOutOfBoundsException {
    getSequenceRef(i).setComments(comments);
  }

  public Sequence getSequence(int i) throws IndexOutOfBoundsException {
    // Specified sequence existence verification
    if (i < _sequences.size()) return (Sequence)((Sequence)_sequences.get(i)).clone();
    throw new ArrayIndexOutOfBoundsException(i);
  }

  public void setSequence(int i, Sequence sequence, boolean checkNames) throws Exception {
    // Sequence's name existence checking
    if (checkNames) {
      // For all names in vector : throw exception if name already exists
      for(int j = 0; j < _sequences.size(); j++) {
        if (getSequenceRef(j).getName().equals(sequence.getName()))
          if (j != i) throw new Exception("VectorSequenceContainer::setSequence : Sequence's name already exists in container");
      }
    }

    // New sequence's alphabet and sequence container's alphabet matching verification
    if (sequence.getAlphabet().getAlphabetType().equals(_alphabet.getAlphabetType())) {
      // New sequence insertion in sequence container
      _sequences.set(i, sequence.clone());
    } else throw new AlphabetMismatchException("VectorSequenceContainer::setSequence : Alphabets don't match", _alphabet, sequence.getAlphabet());
  }

  public Sequence removeSequence(int i) throws IndexOutOfBoundsException {
    // Copy sequence:
    Sequence deleted = getSequenceRef(i);
    // Remove pointer toward old sequence:
    _sequences.remove(i);
    // Send copy:
    return deleted;
  }

  public void deleteSequence(int i) throws IndexOutOfBoundsException {
    // Remove pointer toward old sequence:
    _sequences.remove(i);
  }
  /** @} */

  /**
   * @name Add sequence to this container.
   *
   * @{
   */

  /**
   * @brief Add a sequence at the end of the container.
   *
   * The sequence is copied into the container.
   * If checkNames is set to true, the method check if the name of the
   * sequence is already used in the container, and sends an exception if it
   * is the case. Otherwise, do not check the name: the method is hence faster,
   * but use it at your own risks!
   *
   * @param sequence The sequence to add.
   * @param checkNames Tell if the method must check the name of the sequence
   * before adding it.
   * @throw Exception If the sequence couldn't be added to the container.
   */
  public void addSequence(Sequence sequence, boolean checkNames) throws Exception {
    // Sequence's name existence checking
    if (checkNames) {
      // For all names in vector : throw exception if name already exists
      for(int i = 0; i < _sequences.size(); i++) {
        if (getSequenceRef(i).getName().equals(sequence.getName()))
          throw new Exception("VectorSequenceContainer::addSequence : Sequence's name already exists in container");
      }
    }

    // New sequence's alphabet and sequence container's alphabet matching verification
    if (sequence.getAlphabet().getAlphabetType().equals(_alphabet.getAlphabetType())) {
      _sequences.add(sequence.clone());
    } else throw new AlphabetMismatchException("VectorSequenceContainer::addSequence : Alphabets don't match", _alphabet, sequence.getAlphabet());
  }

  /**
   * @brief Add a sequence to the container at a particular position.
   *
   * The sequence is copied into the container.
   * If checkNames is set to true, the method check if the name of the
   * sequence is already used in the container, and sends an exception if it
   * is the case. Otherwise, do not check the name: the method is hence faster,
   * but use it at your own risks!
   *
   * @param sequence The sequence to add.
   * @param pos      The position where to insert the new sequence.
   * All the following sequences will be pushed.
   * @param checkNames Tell if the method must check the name of the sequence
   * before adding it.
   * @throw Exception If the sequence couldn't be added to the container.
   */
  public void addSequence(Sequence sequence, int pos, boolean checkNames) throws Exception {
    // Sequence's name existence checking
    if (checkNames) {
      // For all names in vector : throw exception if name already exists
      for(int i = 0; i < _sequences.size(); i++) {
        if (getSequenceRef(i).getName().equals(sequence.getName()))
          throw new Exception("VectorSequenceContainer::addSequence : Sequence's name already exists in container");
      }
    }

    // New sequence's alphabet and sequence container's alphabet matching verification
    if (sequence.getAlphabet().getAlphabetType().equals(_alphabet.getAlphabetType())) {
      _sequences.add(pos, sequence.clone());
    } else throw new AlphabetMismatchException("VectorSequenceContainer::addSequence : Alphabets don't match", _alphabet, sequence.getAlphabet());
  }

  /**
   * @name Inner methods.
   *
   * @{
   */
  public Sequence getSequenceRef(int i) throws IndexOutOfBoundsException {
    return (Sequence)_sequences.get(i);
  }

  public Sequence getSequenceRef(String name) throws SequenceNotFoundException {
    int i = getSequencePosition(name);
    return (Sequence)_sequences.get(i);
  }
  /** @} */

}
