package bio.sequences.io;

/**
 * <p>Title: Bio</p>
 * <p>Description: Tools for bioinformatics</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Julien Dutheil
 * @version 1.0
 */

import bio.sequences.*;
import gnu.jtools.utils.file.FileTools;
import gnu.jtools.utils.string.TextTools;
import java.io.*;
import java.util.*;
import javax.swing.filechooser.*;

public class Phylip
    extends AbstractIOSequence {

  protected boolean _extended, _sequential;

  public Phylip(boolean extended, boolean sequential, int charsByLine) {
    super(charsByLine);
    _extended = extended;
    _sequential = sequential;
  }

  public Phylip(boolean extended, boolean sequential) {
    super();
    _extended = extended;
    _sequential = sequential;
  }

  private String[] splitNameAndSequence(String s) {
    String[] v = new String[2];
    if (_extended) {
      int index = s.indexOf("  ");
      v[0] = TextTools.removeFirstAndLastBlanks(s.substring(0, index));
      v[1] = TextTools.removeFirstBlanks(s.substring(index + 2)); //There may be more than 6 white spaces.
    }  else {
      v[0] = TextTools.removeFirstAndLastBlanks(s.substring(0, 10));
      v[1] = s.substring(11);
    }
    return v;
  }

  private void readSequential(BufferedReader br, VectorSequenceContainer vsc) throws
      Exception {
    String temp;

    //Ignore first line:
    br.readLine();
    temp = FileTools.getNextLine(br);

    while (temp != null) {
      // Read each sequence:
      String[] v = splitNameAndSequence(temp);
      String name = v[0];
      String seq = v[1];
      while (!TextTools.isVoid(temp)) {
        //Sequences are separated by at least one blank line:
        temp = br.readLine(); // read next line in file.
        seq += TextTools.removeBlanks(temp);
      }
      //end of this sequence:
      vsc.addSequence(new Sequence(name, seq, vsc.getAlphabet()), true);
      temp = FileTools.getNextLine(br);
    }
  }

  private void readInterleaved(BufferedReader br, VectorSequenceContainer vsc) throws
      Exception {
    String temp;

    //Read first line:
    temp = br.readLine();
    StringTokenizer st = new StringTokenizer(temp);
    int nbSequences = Integer.parseInt(st.nextToken());
    //int nbSites     = Integer.parseInt(st.nextToken());
    temp = FileTools.getNextLine(br);

    List names = new ArrayList(), seqs = new ArrayList();
    // Read first block:
    for (int i = 0; i < nbSequences && temp != null && !TextTools.isVoid(temp); i++) {
      String[] v = splitNameAndSequence(temp);
      names.add(v[0]);
      seqs.add(v[1]);
      temp = br.readLine();
    }

    //Then read all other blocks:
    temp = FileTools.getNextLine(br);
    while (temp != null) {
      for (int i = 0; i < names.size(); i++) {
        //System.out.println(i+"\t"+temp);
        if (TextTools.isVoid(temp)) {
          throw new IOException("Phylip::readInterleaved. Bad file,there are not the same number of sequence in each block.");
        }
        seqs.set(i, ((String)seqs.get(i)) + TextTools.removeBlanks(temp));
        temp = br.readLine(); // read next line in file.
      }
      temp = FileTools.getNextLine(br);
    }
    for (int i = 0; i < names.size(); i++) {
      vsc.addSequence(new Sequence((String)names.get(i), (String)seqs.get(i), vsc.getAlphabet()), true);
    }
  }

// Method to read a mase file, with filling of an existing vector type sequences container
  public void read(File file, VectorSequenceContainer vsc) throws Exception {
    // Checking the existence of specified file
    BufferedReader br = new BufferedReader(new FileReader(file));

    if (_sequential) {
      readSequential(br, vsc);
    } else {
      readInterleaved(br, vsc);
    }
    br.close();
  }

  private String[] getSizedNames(String[] names) {
    String[] sizedNames = new String[names.length];
    if (_extended) {
      //Add 6 white spaces to the larger name and align other names.
      //First, determine the size of the wider name:
      int sizeMax = 0;
      for (int i = 0; i < names.length; i++) {
        if (names[i].length() > sizeMax) {
          sizeMax = names[i].length();
          //Quite easy ;-) Now update all lengths:
        }
      }
      for (int i = 0; i < names.length; i++) {
        sizedNames[i] = TextTools.resizeRight(names[i], sizeMax + 2, ' ');

      }
    }
    else {
      //We trunc all names to ten characters:
      for (int i = 0; i < names.length; i++) {
        sizedNames[i] = TextTools.resizeRight(names[i], 10, ' ');
      }
      System.out.println("Warning: names have been truncated to 10 characters. They may be ambiguous sequence names then.");
    }
    return sizedNames;
  }

  private void writeSequential(PrintWriter pw, SequenceContainer sc,
                               int charsByLine) {
    try {
      int numberOfSites = sc.getSequence(sc.getSequencesNames()[0]).size();
      pw.println(sc.getNumberOfSequences() + " " + numberOfSites);
    } catch(SequenceNotFoundException snfe) {
      System.out.println("This should never happen!!!");
      snfe.printStackTrace();
    }

    String[] seqNames = sc.getSequencesNames();
    String[] names = getSizedNames(seqNames);
    for (int i = 0; i < seqNames.length; i++) {
      try {
        String[] seq = TextTools.split(sc.toString(seqNames[i]), charsByLine);
        pw.println(names[i] + seq[0]);
        for (int j = 1; j < seq.length; j++) {
          pw.println(TextTools.rep(' ', names[i].length()) + seq[j]);
        }
        pw.println();
      } catch(SequenceNotFoundException snfe) {
        System.out.println("This should never happen!!!");
        snfe.printStackTrace();
      }
    }
  }

  private void writeInterleaved(PrintWriter pw, SequenceContainer sc, int charsByLine) {
    //cout << "Write interleaved;" << endl;
    try {
      int numberOfSites = sc.getSequence(sc.getSequencesNames()[0]).size();
      pw.println(sc.getNumberOfSequences() + " " + numberOfSites + " I");
    } catch(SequenceNotFoundException snfe) {
      System.out.println("This should never happen!!!");
      snfe.printStackTrace();
    }

    String[] seqNames = sc.getSequencesNames();
    String[] names = getSizedNames(seqNames);
    //Split sequences:
    String[][] seqs  = new String[sc.getNumberOfSequences()][];
    try {
      for (int i = 0; i < seqNames.length; i++) {
        seqs[i] = TextTools.split(sc.toString(seqNames[i]), charsByLine);
      }
    } catch(SequenceNotFoundException snfe) {
      System.out.println("This should never happen!!!");
      snfe.printStackTrace();
    }
    //Write first block:
    for (int i = 0; i < names.length; i++) {
      pw.println(names[i] + seqs[i][0]);
    }
    pw.println();
    //Write other blocks:
    for (int j = 1; j < seqs[0].length; j++) {
      for (int i = 0; i < sc.getNumberOfSequences(); i++) {
        pw.println(seqs[i][j]);
      }
      pw.println();
    }
  }

  /******************************************************************************/
// Methods to write a sequence container in phylip file
// Specified file will be created if not exists, and else the new sequences will be added at end of file
  public void write(File file, SequenceContainer sc, boolean overwrite) throws Exception {
    //First must check if all sequences are aligned:
    if (sc.getNumberOfSequences() == 0) {
      throw new Exception(
          "Phylip::write. SequenceContainer appear to contain no sequence.");
    }

    if (!SequenceContainerTools.sequencesHaveTheSameLength(sc)) {
      throw new SequenceNotAlignedException(
          "Phylip::write. Sequences have to e of same length.", null);
    }

    // Open file in specified mode
    PrintWriter pw = new PrintWriter(new FileWriter(file, !overwrite));

    if (_sequential) {
      writeSequential(pw, sc, _charsByLine);
    }
    else {
      writeInterleaved(pw, sc, _charsByLine);

    }
    pw.close();
  }

  public String getFormatName() {
    return "Phylip file, " + (_extended ? "extended," : "") +
        (_sequential ? "sequential" : "interleaved");
  }

  public String getFormatDescription() {
    return
        "Phylip file format, sequential and interleaved. PAML extension also supported.";
  }

  public static javax.swing.filechooser.FileFilter PHYLIP_FILTER = new javax.
      swing.filechooser.FileFilter() {

    public boolean accept(java.io.File file) {
      if (file.isDirectory()) {
        return true;
      }
      String extension = FileTools.getExtension(file);
      if (extension != null) {
        if (extension.equals("phylip") ||
            extension.equals("phy")) {
          return true;
        }
        else {
          return false;
        }
      }
      return false;
    }

    // The description of this filter
    public String getDescription() {
      return "Phylip files (*.phylip, *.phy)";
    }
  };
}
