// -------------------------------------------------------------------------
//
// Copyright or © or Copr. CNRS, (November 16, 2004)
//
// This software is a library providing tool for bioinformatics.
//
// This software is governed by the CeCILL  license under French law and
// abiding by the rules of distribution of free software.  You can  use,
// modify and/ or redistribute the software under the terms of the CeCILL
// license as circulated by CEA, CNRS and INRIA at the following URL
// "http://www.cecill.info".
//
// As a counterpart to the access to the source code and  rights to copy,
// modify and redistribute granted by the license, users are provided only
// with a limited warranty  and the software's author,  the holder of the
// economic rights,  and the successive licensors  have only  limited
// liability.
//
// In this respect, the user's attention is drawn to the risks associated
// with loading,  using,  modifying and/or developing or reproducing the
// software by the user in light of its specific status of free software,
// that may mean  that it is complicated to manipulate,  and  that  also
// therefore means  that it is reserved for developers  and  experienced
// professionals having in-depth computer knowledge. Users are therefore
// encouraged to load and test the software's suitability as regards their
// requirements in conditions enabling the security of their systems and/or
// data to be ensured and,  more generally, to use and operate it in the
// same conditions as regards security.
//
// The fact that you are presently reading this means that you have had
// knowledge of the CeCILL license and that you accept its terms.
//
// -------------------------------------------------------------------------

package bio.sequences.xml;

/**
 * <p>Titre : Sequences Operating System</p>
 * <p>Description : an Operating System dedicated to sequence files analysis.</p>
 * <p>Copyright : Copyright (c) Julien Dutheil</p>
 * @author Julien Dutheil
 * @version 1.0
 */

import bio.*;
import bio.sequences.*;
import gnu.jtools.utils.attributes.Attribute;
import gnu.jtools.utils.attributes.AttributesMap;
import java.util.*;

public class XSequence extends Sequence {

  protected AttributesMap _attributes;

  public static final String SEQUENCE          = "Sequence";
  public static final String SEQUENCES_BLOCK   = "SequencesBloc";
  public static final String SEQUENCE_CHAIN    = "Sequence chain";
  public static final String SEQUENCE_NAME     = "Sequence name";
  public static final String SEQUENCE_COMMENTS = "Sequence comments";

  /**
   * @brief Direct constructor: build a Sequence object from a string
   * You can use it safely for RNA, DNA and protein sequences.
   *
   * It can be used with codon sequences too, the string will be cut into
   * parts of size 3. But for more complicated alphabets, you should use one
   * complete constructors.
   *
   * @param name     The sequence name.
   * @param sequence The whole sequence to be parsed as a string.
   * @param alpha    A pointer toward the alphabet to be used with this sequence.
   */
  public XSequence(String name, String sequence, Alphabet alpha) throws BadCharException {
    super(name, sequence, alpha);
    _attributes = new AttributesMap();
  }

  /**
   * @brief Direct constructor: build a Sequence object from a string
   * You can use it safely for RNA, DNA and protein sequences.
   *
   * It can be used with codon sequences too, the string will be cut into
   * parts of size 3. But for more complicated alphabets, you should use one
   * complete constructors.
   *
   * @param name     The sequence name.
   * @param sequence The whole sequence to be parsed as a string.
   * @param comments Comments to add to the sequence.
   * @param alpha    A pointer toward the alphabet to be used with this sequence.
   */
  public XSequence(String name, String sequence, List comments, Alphabet alpha) throws BadCharException {
    super(name, sequence, comments, alpha);
    _attributes = new AttributesMap();
  }

  /**
   * @brief General purpose constructor, can be used with any alphabet.
   *
   * You should note that the sequence is stored as a vector of int.
   * Hence each string in the vector will be translated using the alphabet object.
   *
   * @param name     The sequence name.
   * @param sequence The sequence content.
   * @param alpha    A pointer toward the alphabet to be used with this sequence.
   */
  public XSequence(String name, String[] sequence, Alphabet alpha) throws BadCharException {
    super(name, sequence, alpha);
    _attributes = new AttributesMap();
  }

  /**
   * @brief General purpose constructor, can be used with any alphabet.
   *
   * You should note that the sequence is stored as a vector of int.
   * Hence each string in the vector will be translated using the alphabet object.
   *
   * @param name     The sequence name.
   * @param sequence The sequence content.
   * @param comments Comments to add to the sequence.
   * @param alpha    A pointer toward the alphabet to be used with this sequence.
   */
  public XSequence(String name, String[] sequence, List comments, Alphabet alpha) throws BadCharException {
    super(name, sequence, comments, alpha);
    _attributes = new AttributesMap();
  }

  /**
   * @brief General purpose constructor, can be used with any alphabet.
   *
   * @param name     The sequence name.
   * @param sequence The sequence content.
   * @param alpha    A pointer toward the alphabet to be used with this sequence.
   */
  public XSequence(String name, Vint sequence, Alphabet alpha) throws BadIntException {
    super(name, sequence, alpha);
    _attributes = new AttributesMap();
  }

  /**
   * @brief General purpose constructor, can be used with any alphabet.
   *
   * @param name     The sequence name.
   * @param sequence The sequence content.
   * @param comments Comments to add to the sequence.
   * @param alpha    A pointer toward the alphabet to be used with this sequence.
   */
  public XSequence(String name, Vint sequence, List comments, Alphabet alpha) throws BadIntException {
    super(name, sequence, comments, alpha);
    _attributes = new AttributesMap();
  }

  /**
   * @brief Build a new XSequence from a Sequence object.
   */
  public XSequence(Sequence s) {
    super(s);
    _attributes = new AttributesMap();
  }

  /**
   * @brief The Sequence copy constructor. This does not perform a hard copy of the alphabet object.
   */
  public XSequence(XSequence s) {
    super(s);
    _attributes = new AttributesMap(s.getAttributes());
  }

  /**
  * @name The Cloneable interface
  *
  * @{
  */
  public Object clone() {
    return new XSequence(this);
  }



  /**
   * Attributes:
   */

  /**
   * Retrieve attributes:
   */

  public Object getAttribute(String name) {
    return _attributes.get(name);
  }

  public AttributesMap getAttributes() {
    return _attributes;
  }

  /**
   * Set attributes:
   */

  public Attribute setAttribute(Attribute attribute) {
    return (Attribute)_attributes.put(attribute);
  }

  public Attribute setAttribute(String name, Object value) {
    return (Attribute)_attributes.put(name, value);
  }

  public Attribute setAttribute(String name, Object value, boolean isNameEditable, boolean isValueEditable) {
    return (Attribute)_attributes.put(name, value, isNameEditable, isValueEditable);
  }

  public void setAttributes(AttributesMap attributes) {
    _attributes = attributes;
  }

  public void setAttributeNameEditable(String name, boolean isNameEditable) {
    try { _attributes.getAttribute(name).setNameEditable(isNameEditable); }
    catch(NullPointerException e) {} //No attribute with this name, operation cancelled.
  }

  public void setAttributeValueEditable(String name, boolean isValueEditable) {
    try { _attributes.getAttribute(name).setValueEditable(isValueEditable); }
    catch(NullPointerException e) {} //No attribute with this name, operation cancelled.
  }

  public void appendAttributeStringValue(String name, String value, String delimiter) {
    //System.out.println(name + "\t" + value);
    Attribute a = _attributes.getAttribute(name);
    if(a == null) {//Create a new attribute:
      _attributes.put(name, value);
    } else {
      if(a.getValue().getClass() != String.class) {
        System.err.println("Attribute " + name + " has not a string value. Can't append the value '" + value + "'.");
      } else {
        a.setValue(((String)a.getValue()) + delimiter + value);
      }
    }
  }

  /**
   * Remove attributes
   */

  public Attribute removeAttribute(Attribute attribute) {
    return _attributes.remove(attribute);
  }

  public Attribute removeAttribute(String name) {
    return _attributes.remove(name);
  }

  /**
   * Get information about attributes:
   */

  public int getNumberOfAttributes() {
    return _attributes.size();
  }

  public boolean hasAttribute(String name) {
    return getAttribute(name) != null;
  }
}
