package gnu.jtools.stats.csv;

/**
 * <p>Title: Stats</p>
 * <p>Description: Statistics tools</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: </p>
 * @author Julien Dutheil
 * @version 1.0
 */

import gnu.jtools.stats.table.InvalidTableColumnSizeException;
import gnu.jtools.stats.table.InvalidTableRowSizeException;
import gnu.jtools.stats.table.Table;
import gnu.jtools.utils.file.FileTools;
import gnu.jtools.utils.gui.GraphicsTools;
import gnu.jtools.utils.string.TextTools;

import java.io.*;
import java.util.*;
import javax.swing.*;
import java.awt.event.*;
import java.awt.Frame;
import java.awt.GridLayout;

public class CSVIO {

  /**
   * Build a new CSVIO object.
   */
  public CSVIO(String na) {
    this.delimiter = ',';
    this.decimal   = '.';
    this.quote     = '\"';
    this.numberPolicy = ALL_DOUBLES;
    this.na = "NA";
  }

  /**
   * Read a table from a CSV file.
   *
   * @param file     The file to parse.
   * @param header   Tell if the file contain a header line.
   * @param rowNames Tell if the first column must be used as row names.
   * @return A Table object.
   */
  public Table readTable(File file, boolean header, boolean rowNames) throws IOException {

    BufferedReader reader = new BufferedReader(new FileReader(file));
    String line = reader.readLine();

    // Analyse header:
    List cNames = null;
    if(header) {
      cNames = parseLine(line);
      //check if first cell is void:
      if(TextTools.isVoid(cNames.get(0).toString())) cNames.remove(0);
    }

    // The first line determines the types to use:
    line = reader.readLine();
    List row = parseLine(line);

    // Row names:
    List rNames = new ArrayList();

    if(rowNames) rNames.add(row.remove(0));
    if(header && row.size() != cNames.size()) throw new IOException("All rows must have the same length (error at line 1 in data file)!");
    Class[] types = new Class[row.size()];
    for(int j = 0; j < row.size(); j++) types[j] = row.get(j).getClass();

    // Builds a new table objects with these types:
    Table table = new Table(types);
    // Add the first line:
    try {
      table.addRow(row);
    } catch(InvalidTableRowSizeException itrse) {
      // This can't happen!
    }

    line = reader.readLine();
    int rowCount = 2;
    while(line != null) {
      row = parseLine(line);
      if(rowNames) rNames.add(row.remove(0));
      try { table.addRow(row); }
      catch(ClassCastException cce) {
        throw new IOException("Each column must be of one single type (error at line " + rowCount + " in data file).");
      } catch(InvalidTableRowSizeException itrse) {
        throw new IOException("All rows must have the same length (error at line " + rowCount + " in data file)!");
      }
      // Read next line:
      line = reader.readLine();
      rowCount++;
    }

    // Set column names:
    try {
      if (header)
        //for(int i = 0; i < cNames.size(); i++) System.out.println(cNames.get(i).getClass().toString() + "\t" + cNames.get(i).toString());
        table.setColumnNames( (String[]) cNames.toArray(new String[0]));
    } catch(InvalidTableRowSizeException itcse) {
      itcse.printStackTrace();
      throw new IOException("Unexpected error while loading table.");
    }
    // Set row names:
    try {
      if (rowNames)
        table.setRowNames( (String[]) rNames.toArray(new String[0]));
    } catch(InvalidTableColumnSizeException itrse) {
      itrse.printStackTrace();
      throw new IOException("Unexpected error while loading table.");
    }

    return table;
  }

  public  void writeTable(Table table, PrintWriter pw) {
    int n = table.getNumberOfColumns();
    if(table.hasColumnNames() && n > 0) { // Print column names:
      for(int j = 0; j < n - 1; j++)
        pw.print(table.getColumnName(j) + delimiter);
      pw.println(table.getColumnName(n - 1));
    }

    for(int i = 0; i < table.getNumberOfRows(); i++) { // for each row
      if (table.hasRowNames()) pw.print(table.getRowName(i) + delimiter);
      if (n > 0) {
        for (int j = 0; j < n - 1; j++) { // for each column
          Object o = table.getCell(i, j);
          pw.print(o == null ? na : toText(o.toString()) + delimiter);
        }
        // last column
        Object o = table.getCell(i, n - 1);
        pw.println(o == null ? na : toText(o.toString()));
      }
    }
  }

  protected Object getCell(String content) {
    //System.out.println(content);
    if(content.equalsIgnoreCase(na)) return null;
    if(numberPolicy == ALL_STRINGS) return content;
    if(numberPolicy == ALLOW_INTEGERS)
      try { return new Integer(content); } catch(NumberFormatException nfe) {}
    try { return new Double (content); } catch(NumberFormatException nfe) {}
    if(content.equalsIgnoreCase("true") || content.equalsIgnoreCase("false"))
      try { return new Boolean(content); } catch(NumberFormatException nfe) {}
    return content;
  }

  protected List parseLine(String line) {
    List list = new ArrayList();
    int currentPos = 0;
    boolean quoteOpened = false;
    loop:for (int i = 0; i < line.length(); i++) {
      //analyse each char:
      if (line.charAt(i) == quote) {
        if (quoteOpened) {
          //A quote is already opened, so this is an ending quote...
          quoteOpened = false;
          //Now we can add the content of the string to the list:
          list.add(line.substring(currentPos, i));
          //System.out.println("Add " + line.substring(currentPos, i) + " 'cause I found a closing quote.");
          //Jump to next field:
          int next = line.indexOf(delimiter, i + 1);
          if (next != -1) {
            i = next; //i will be incremented!
            currentPos = i + 1;
            if (currentPos == line.length()) {
              //the last field is void, must add it now:
              list.add("");
            }
          }
          else {
            //this is the last field
            currentPos = line.length(); //Current position is set to the end of the line.
          }
        }
        else {
          //No quote is opened, so this one is an opening one!
          quoteOpened = true;
          currentPos = i + 1; //Move the currentPosition to the beginning of the text.
        }
      }
      else if (line.charAt(i) == delimiter) {
        if (quoteOpened)
          continue loop; //consider this char as part of the text.
        list.add(getCell(line.substring(currentPos, i)));
        //System.out.println("Add " + line.substring(currentPos, i) + " 'cause I found a field delimiter.");
        currentPos = i + 1;
        if (currentPos == line.length()) {
          //the last field is void, must add it now:
          list.add("");
        }
      }
    }
    //Now add last field if there is one:
    if (currentPos < line.length()) {
      list.add(getCell(line.substring(currentPos)));
    }
    return list;
  }

  private String toText(Object o) {
    if(o == null) return "" + quote + quote;
    if(!numberPolicy.equals(ALL_STRINGS) && o.getClass().isAssignableFrom(Number.class)) return o.toString();
    return quote + TextTools.removeEOL(o.toString()) + quote;
  }

  static class OptionDialog extends JDialog {

    public OptionDialog(Frame owner) {
      super(owner, "CSV Parameters", true);
      state = OK;
      initGUI();
      this.addWindowListener(new WindowAdapter() {
        public void windowClosing(WindowEvent e) {
          state = CANCEL;
        }
      });
      pack();
      GraphicsTools.center(this);
    }

    private void initGUI() {
      fieldDelimiter = new JComboBox(defaultFieldDelimiters);
      fieldDelimiter.setEditable(true);
      textDelimiter = new JComboBox(defaultTextDelimiters);
      textDelimiter.setEditable(true);
      ok = new JButton("Ok");
      ok.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          state = OK;
          setVisible(false);
        }
      });
      getContentPane().setLayout(new GridLayout(3, 2));
      getContentPane().add(new JLabel("Field delimiter:"));
      getContentPane().add(fieldDelimiter);
      getContentPane().add(new JLabel("Text delimiter:"));
      getContentPane().add(textDelimiter);
      getContentPane().add(ok);
    }

    private JComboBox fieldDelimiter, textDelimiter;
    private static final String[] defaultFieldDelimiters = {"\t(tabulation)", ",", ";"}, defaultTextDelimiters = {"\"", "\'"};
    private JButton ok;
    private int state;
    public static final int OK = 0, CANCEL = -1;
  }

  public int showOptionDialog(Frame owner) {
    optionDialog = new OptionDialog(owner);
    optionDialog.setVisible(true);
    int test = optionDialog.state;
    if(test == optionDialog.OK) {
      //Actualize parameters:
      delimiter = ((String)optionDialog.fieldDelimiter.getSelectedItem()).charAt(0);
      quote     = ((String)optionDialog. textDelimiter.getSelectedItem()).charAt(0);
    }
    return test;
  }

  public void setNumberPolicy(String numberPolicy) { this.numberPolicy = numberPolicy; }

  public String getNumberPolicy() { return numberPolicy; }

  public static final javax.swing.filechooser.FileFilter CSV_FILTER = new javax.swing.filechooser.FileFilter() {

    public boolean accept(File file) {
      if(file.isDirectory()) return true;
      String extension = FileTools.getExtension(file);
      if(extension != null) {
        if(extension.equals("csv")) return true;
        else return false;
      }
      return false;
    }

    // The description of this filter
    public String getDescription() {
      return "CSV (*.csv)";
    }
  };

  protected char delimiter, decimal, quote;
  protected OptionDialog optionDialog;
  protected String numberPolicy;
  protected String na;

  public static final String ALLOW_INTEGERS = "allow integers";
  public static final String ALL_DOUBLES    = "all doubles";
  public static final String ALL_STRINGS    = "all strings";
}
